# -*- coding: utf-8 -*-
# File: common.py

from __future__ import division
import itertools
import numpy as np
import pprint
from collections import defaultdict, deque
from copy import copy
import six
import tqdm
from termcolor import colored

from ..utils import logger
from ..utils.utils import get_rng, get_tqdm, get_tqdm_kwargs
from .base import DataFlow, DataFlowReentrantGuard, ProxyDataFlow, RNGDataFlow

try:
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping

__all__ = ['TestDataSpeed', 'PrintData', 'BatchData', 'BatchDataByShape', 'FixedSizeData', 'MapData',
           'MapDataComponent', 'RepeatedData', 'RepeatedDataPoint', 'RandomChooseData',
           'RandomMixData', 'JoinData', 'ConcatData', 'SelectComponent',
           'LocallyShuffleData', 'CacheData']


class TestDataSpeed(ProxyDataFlow):
    """ Test the speed of a DataFlow """
    def __init__(self, ds, size=5000, warmup=0):
        """
        Args:
            ds (DataFlow): the DataFlow to test.
            size (int): number of datapoints to fetch.
            warmup (int): warmup iterations
        """
        super(TestDataSpeed, self).__init__(ds)
        self.test_size = int(size)
        self.warmup = int(warmup)
        self._reset_called = False

    def reset_state(self):
        self._reset_called = True
        super(TestDataSpeed, self).reset_state()

    def __iter__(self):
        """ Will run testing at the beginning, then produce data normally. """
        self.start()
        yield from self.ds

    def start(self):
        """
        Start testing with a progress bar.
        """
        if not self._reset_called:
            self.ds.reset_state()
        itr = self.ds.__iter__()
        if self.warmup:
            for _ in tqdm.trange(self.warmup, **get_tqdm_kwargs()):
                next(itr)
        # add smoothing for speed benchmark
        with get_tqdm(total=self.test_size,
                      leave=True, smoothing=0.2) as pbar:
            for idx, dp in enumerate(itr):
                pbar.update()
                if idx == self.test_size - 1:
                    break


class BatchData(ProxyDataFlow):
    """
    Stack datapoints into batches.
    It produces datapoints of the same number of components as ``ds``, but
    each component has one new extra dimension of size ``batch_size``.
    The batch can be either a list of original components, or (by default)
    a numpy array of original components.
    """

    def __init__(self, ds, batch_size, remainder=False, use_list=False):
        """
        Args:
            ds (DataFlow): A dataflow that produces either list or dict.
                When ``use_list=False``, the components of ``ds``
                must be either scalars or :class:`np.ndarray`, and have to be consistent in shapes.
            batch_size(int): batch size
            remainder (bool): When the remaining datapoints in ``ds`` is not
                enough to form a batch, whether or not to also produce the remaining
                data as a smaller batch.
                If set to False, all produced datapoints are guaranteed to have the same batch size.
                If set to True, `len(ds)` must be accurate.
            use_list (bool): if True, each component will contain a list
                of datapoints instead of an numpy array of an extra dimension.
        """
        super(BatchData, self).__init__(ds)
        if not remainder:
            try:
                assert batch_size <= len(ds)
            except NotImplementedError:
                pass
        self.batch_size = int(batch_size)
        assert self.batch_size > 0
        self.remainder = remainder
        self.use_list = use_list

    def __len__(self):
        ds_size = len(self.ds)
        div = ds_size // self.batch_size
        rem = ds_size % self.batch_size
        if rem == 0:
            return div
        return div + int(self.remainder)

    def __iter__(self):
        """
        Yields:
            Batched data by stacking each component on an extra 0th dimension.
        """
        holder = []
        for data in self.ds:
            holder.append(data)
            if len(holder) == self.batch_size:
                yield BatchData.aggregate_batch(holder, self.use_list)
                del holder[:]
        if self.remainder and len(holder) > 0:
            yield BatchData.aggregate_batch(holder, self.use_list)

    @staticmethod
    def _batch_numpy(data_list):
        data = data_list[0]
        if isinstance(data, six.integer_types):
            dtype = 'int32'
        elif type(data) == bool:
            dtype = 'bool'
        elif type(data) == float:
            dtype = 'float32'
        elif isinstance(data, (six.binary_type, six.text_type)):
            dtype = 'str'
        else:
            try:
                dtype = data.dtype
            except AttributeError:
                raise TypeError("Unsupported type to batch: {}".format(type(data)))
        try:
            return np.asarray(data_list, dtype=dtype)
        except Exception as e:  # noqa
            logger.exception("Cannot batch data. Perhaps they are of inconsistent shape?")
            if isinstance(data, np.ndarray):
                s = pprint.pformat([x.shape for x in data_list])
                logger.error("Shape of all arrays to be batched: " + s)
            try:
                # open an ipython shell if possible
                import IPython as IP; IP.embed()    # noqa
            except ImportError:
                pass

    @staticmethod
    def aggregate_batch(data_holder, use_list=False):
        """
        Aggregate a list of datapoints to one batched datapoint.

        Args:
            data_holder (list[dp]): each dp is either a list or a dict.
            use_list (bool): whether to batch data into a list or a numpy array.

        Returns:
            dp:
                either a list or a dict, depend on the inputs.
                Each item is a batched version of the corresponding inputs.
        """
        first_dp = data_holder[0]
        if isinstance(first_dp, (list, tuple)):
            result = []
            for k in range(len(first_dp)):
                data_list = [x[k] for x in data_holder]
                if use_list:
                    result.append(data_list)
                else:
                    result.append(BatchData._batch_numpy(data_list))
        elif isinstance(first_dp, dict):
            result = {}
            for key in first_dp.keys():
                data_list = [x[key] for x in data_holder]
                if use_list:
                    result[key] = data_list
                else:
                    result[key] = BatchData._batch_numpy(data_list)
        else:
            raise ValueError("Data point has to be list/tuple/dict. Got {}".format(type(first_dp)))
        return result


class BatchDataByShape(BatchData):
    """
    Group datapoints of the same shape together to batches.
    It doesn't require input DataFlow to be homogeneous anymore: it can have
    datapoints of different shape, and batches will be formed from those who
    have the same shape.

    Note:
        It is implemented by a dict{shape -> datapoints}.
        Therefore, datapoints of uncommon shapes may never be enough to form a batch and
        never get generated.
    """
    def __init__(self, ds, batch_size, idx):
        """
        Args:
            ds (DataFlow): input DataFlow. ``dp[idx]`` has to be an :class:`np.ndarray`.
            batch_size (int): batch size
            idx (int): ``dp[idx].shape`` will be used to group datapoints.
                Other components are assumed to be batch-able.
        """
        super(BatchDataByShape, self).__init__(ds, batch_size, remainder=False)
        self.idx = idx

    def reset_state(self):
        super(BatchDataByShape, self).reset_state()
        self.holder = defaultdict(list)
        self._guard = DataFlowReentrantGuard()

    def __iter__(self):
        with self._guard:
            for dp in self.ds:
                shp = dp[self.idx].shape
                holder = self.holder[shp]
                holder.append(dp)
                if len(holder) == self.batch_size:
                    yield BatchData.aggregate_batch(holder)
                    del holder[:]


class FixedSizeData(ProxyDataFlow):
    """ Generate data from another DataFlow, but with a fixed total count.
    """
    def __init__(self, ds, size, keep_state=True):
        """
        Args:
            ds (DataFlow): input dataflow
            size (int): size
            keep_state (bool): keep the iterator state of ``ds``
                between calls to :meth:`__iter__()`, so that the
                next call will continue the previous iteration over ``ds``,
                instead of reinitializing an iterator.

        Example:

        .. code-block:: none

            ds produces: 1, 2, 3, 4, 5; 1, 2, 3, 4, 5; ...
            FixedSizeData(ds, 3, True): 1, 2, 3; 4, 5, 1; 2, 3, 4; ...
            FixedSizeData(ds, 3, False): 1, 2, 3; 1, 2, 3; ...
            FixedSizeData(ds, 6, False): 1, 2, 3, 4, 5, 1; 1, 2, 3, 4, 5, 1;...
        """
        super(FixedSizeData, self).__init__(ds)
        self._size = int(size)
        self.itr = None
        self._keep = keep_state

    def __len__(self):
        return self._size

    def reset_state(self):
        super(FixedSizeData, self).reset_state()
        self.itr = self.ds.__iter__()
        self._guard = DataFlowReentrantGuard()

    def __iter__(self):
        with self._guard:
            if self.itr is None:
                self.itr = self.ds.__iter__()
            cnt = 0
            while True:
                try:
                    dp = next(self.itr)
                except StopIteration:
                    self.itr = self.ds.__iter__()
                    dp = next(self.itr)

                cnt += 1
                yield dp
                if cnt == self._size:
                    if not self._keep:
                        self.itr = None
                    return


class MapData(ProxyDataFlow):
    """
    Apply a mapper/filter on the datapoints of a DataFlow.

    Note:
        1. Please make sure func doesn't modify its arguments in place,
           unless you're certain it's safe.
        2. If you discard some datapoints, ``len(MapData(ds))`` will be incorrect.

    Example:

        .. code-block:: none

            ds = Mnist('train')  # each datapoint is [img, label]
            ds = MapData(ds, lambda dp: [dp[0] * 255, dp[1]])
    """

    def __init__(self, ds, func):
        """
        Args:
            ds (DataFlow): input DataFlow
            func (datapoint -> datapoint | None): takes a datapoint and returns a new
                datapoint. Return None to discard/skip this datapoint.
        """
        super(MapData, self).__init__(ds)
        self.func = func

    def __iter__(self):
        for dp in self.ds:
            ret = self.func(copy(dp))  # shallow copy the list
            if ret is not None:
                yield ret


class MapDataComponent(MapData):
    """
    Apply a mapper/filter on a datapoint component.

    Note:
        1. This dataflow itself doesn't modify the datapoints.
           But please make sure func doesn't modify its arguments in place,
           unless you're certain it's safe.
        2. If you discard some datapoints, ``len(MapDataComponent(ds, ..))`` will be incorrect.

    Example:

        .. code-block:: none

            ds = Mnist('train')  # each datapoint is [img, label]
            ds = MapDataComponent(ds, lambda img: img * 255, 0)  # map the 0th component
    """
    def __init__(self, ds, func, index=0):
        """
        Args:
            ds (DataFlow): input DataFlow which produces either list or dict.
            func (TYPE -> TYPE|None): takes ``dp[index]``, returns a new value for ``dp[index]``.
                Return None to discard/skip this datapoint.
            index (int or str): index or key of the component.
        """
        self._index = index
        self._func = func
        super(MapDataComponent, self).__init__(ds, self._mapper)

    def _mapper(self, dp):
        r = self._func(dp[self._index])
        if r is None:
            return None
        dp = copy(dp)   # shallow copy to avoid modifying the datapoint
        if isinstance(dp, tuple):
            dp = list(dp)  # to be able to modify it in the next line
        dp[self._index] = r
        return dp


class RepeatedData(ProxyDataFlow):
    """ Take data points from another DataFlow and produce them until
        it's exhausted for certain amount of times. i.e.:
        dp1, dp2, .... dpn, dp1, dp2, ....dpn
    """

    def __init__(self, ds, num):
        """
        Args:
            ds (DataFlow): input DataFlow
            num (int): number of times to repeat ds.
                Set to -1 to repeat ``ds`` infinite times.
        """
        self.num = num
        super(RepeatedData, self).__init__(ds)

    def __len__(self):
        """
        Raises:
            :class:`ValueError` when num == -1.
        """
        if self.num == -1:
            raise NotImplementedError("__len__() is unavailable for infinite dataflow")
        return len(self.ds) * self.num

    def __iter__(self):
        if self.num == -1:
            while True:
                yield from self.ds
        else:
            for _ in range(self.num):
                yield from self.ds


class RepeatedDataPoint(ProxyDataFlow):
    """ Take data points from another DataFlow and produce them a
    certain number of times. i.e.:
    dp1, dp1, ..., dp1, dp2, ..., dp2, ...
    """

    def __init__(self, ds, num):
        """
        Args:
            ds (DataFlow): input DataFlow
            num (int): number of times to repeat each datapoint.
        """
        self.num = int(num)
        assert self.num >= 1, self.num
        super(RepeatedDataPoint, self).__init__(ds)

    def __len__(self):
        return len(self.ds) * self.num

    def __iter__(self):
        for dp in self.ds:
            for _ in range(self.num):
                yield dp


class RandomChooseData(RNGDataFlow):
    """
    Randomly choose from several DataFlow.
    Stop producing when any of them is exhausted.
    """

    def __init__(self, df_lists):
        """
        Args:
            df_lists (list): a list of DataFlow, or a list of (DataFlow, probability) tuples.
                Probabilities must sum to 1 if used.
        """
        super(RandomChooseData, self).__init__()
        if isinstance(df_lists[0], (tuple, list)):
            assert sum(v[1] for v in df_lists) == 1.0
            self.df_lists = df_lists
        else:
            prob = 1.0 / len(df_lists)
            self.df_lists = [(k, prob) for k in df_lists]

    def reset_state(self):
        super(RandomChooseData, self).reset_state()
        for d in self.df_lists:
            if isinstance(d, tuple):
                d[0].reset_state()
            else:
                d.reset_state()

    def __iter__(self):
        itrs = [v[0].__iter__() for v in self.df_lists]
        probs = np.array([v[1] for v in self.df_lists])
        try:
            while True:
                itr = self.rng.choice(itrs, p=probs)
                yield next(itr)
        except StopIteration:
            return


class RandomMixData(RNGDataFlow):
    """
    Perfectly mix datapoints from several DataFlow using their
    :meth:`__len__()`. Will stop when all DataFlow exhausted.
    """

    def __init__(self, df_lists):
        """
        Args:
            df_lists (list): a list of DataFlow.
                All DataFlow must implement ``__len__()``.
        """
        super(RandomMixData, self).__init__()
        self.df_lists = df_lists
        self.sizes = [len(k) for k in self.df_lists]

    def reset_state(self):
        super(RandomMixData, self).reset_state()
        for d in self.df_lists:
            d.reset_state()

    def __len__(self):
        return sum(self.sizes)

    def __iter__(self):
        sums = np.cumsum(self.sizes)
        idxs = np.arange(self.__len__())
        self.rng.shuffle(idxs)
        idxs = np.array(list(map(
            lambda x: np.searchsorted(sums, x, 'right'), idxs)))
        itrs = [k.__iter__() for k in self.df_lists]
        assert idxs.max() == len(itrs) - 1, "{}!={}".format(idxs.max(), len(itrs) - 1)
        for k in idxs:
            yield next(itrs[k])
        # TODO run till exception


class ConcatData(DataFlow):
    """
    Concatenate several DataFlow.
    Produce datapoints from each DataFlow and start the next when one
    DataFlow is exhausted.
    """

    def __init__(self, df_lists):
        """
        Args:
            df_lists (list): a list of DataFlow.
        """
        self.df_lists = df_lists

    def reset_state(self):
        for d in self.df_lists:
            d.reset_state()

    def __len__(self):
        return sum(len(x) for x in self.df_lists)

    def __iter__(self):
        for d in self.df_lists:
            yield from d


class JoinData(DataFlow):
    """
    Join the components from each DataFlow. See below for its behavior.

    Note that you can't join a DataFlow that produces lists with one that produces dicts.

    Example:

    .. code-block:: none

        df1 produces: [c1, c2]
        df2 produces: [c3, c4]
        joined: [c1, c2, c3, c4]

        df1 produces: {"a":c1, "b":c2}
        df2 produces: {"c":c3}
        joined: {"a":c1, "b":c2, "c":c3}
    """

    def __init__(self, df_lists):
        """
        Args:
            df_lists (list): a list of DataFlow.
                When these dataflows have different sizes, JoinData will stop when any
                of them is exhausted.
                The list could contain the same DataFlow instance more than once,
                but note that in that case `__iter__` will then also be called many times.
        """
        self.df_lists = df_lists

        try:
            self._size = len(self.df_lists[0])
            for d in self.df_lists:
                assert len(d) == self._size, \
                    "All DataFlow must have the same size! {} != {}".format(len(d), self._size)
        except Exception:
            logger.info("[JoinData] Size check failed for the list of dataflow to be joined!")

    def reset_state(self):
        for d in set(self.df_lists):
            d.reset_state()

    def __len__(self):
        """
        Return the minimum size among all.
        """
        return min(len(k) for k in self.df_lists)

    def __iter__(self):
        itrs = [k.__iter__() for k in self.df_lists]
        try:
            while True:
                all_dps = [next(itr) for itr in itrs]
                if isinstance(all_dps[0], (list, tuple)):
                    dp = list(itertools.chain(*all_dps))
                else:
                    dp = {}
                    for x in all_dps:
                        dp.update(x)
                yield dp
        except StopIteration:   # some of them are exhausted
            pass


def SelectComponent(ds, idxs):
    """
    Select / reorder components from datapoints.

    Args:
        ds (DataFlow): input DataFlow.
        idxs (list[int] or list[str]): a list of component indices/keys.

    Example:

    .. code-block:: none

        original df produces: [c1, c2, c3]
        idxs: [2,1]
        this df: [c3, c2]
    """
    return MapData(ds, lambda dp: [dp[i] for i in idxs])


class LocallyShuffleData(ProxyDataFlow, RNGDataFlow):
    """ Buffer the datapoints from a given dataflow, and shuffle them before producing them.
        This can be used as an alternative when a complete random shuffle is too expensive
        or impossible for the data source.

        This dataflow has the following behavior:

        1. It takes datapoints from the given dataflow `ds` to an internal buffer of fixed size.
           Each datapoint is duplicated for `num_reuse` times.
        2. Once the buffer is full, this dataflow starts to yield data from the beginning of the buffer,
           and new datapoints will be added to the end of the buffer. This is like a FIFO queue.
        3. The internal buffer is shuffled after every `shuffle_interval` datapoints that come from `ds`.

        To maintain shuffling states, this dataflow is not reentrant.

        Datapoints from one pass of `ds` will get mixed with datapoints from a different pass.
        As a result, the iterator of this dataflow will run indefinitely
        because it does not make sense to stop the iteration anywhere.
    """

    def __init__(self, ds, buffer_size, num_reuse=1, shuffle_interval=None):
        """
        Args:
            ds (DataFlow): input DataFlow.
            buffer_size (int): size of the buffer.
            num_reuse (int): duplicate each datapoints several times into the buffer to improve
                speed, but duplication may hurt your model.
            shuffle_interval (int): shuffle the buffer after this many
                datapoints were produced from the given dataflow. Frequent shuffle on large buffer
                may affect speed, but infrequent shuffle may not provide enough randomness.
                Defaults to buffer_size / 3
        """
        ProxyDataFlow.__init__(self, ds)
        self.q = deque(maxlen=buffer_size)
        if shuffle_interval is None:
            shuffle_interval = int(buffer_size // 3)
        self.shuffle_interval = shuffle_interval
        self.num_reuse = num_reuse
        self._inf_ds = RepeatedData(ds, -1)

    def reset_state(self):
        self._guard = DataFlowReentrantGuard()
        ProxyDataFlow.reset_state(self)
        RNGDataFlow.reset_state(self)
        self._iter_cnt = 0
        self._inf_iter = iter(self._inf_ds)

    def __len__(self):
        return len(self.ds) * self.num_reuse

    def __iter__(self):
        with self._guard:
            for dp in self._inf_iter:
                self._iter_cnt = (self._iter_cnt + 1) % self.shuffle_interval
                # fill queue
                if self._iter_cnt == 0:
                    self.rng.shuffle(self.q)
                for _ in range(self.num_reuse):
                    if self.q.maxlen == len(self.q):
                        yield self.q.popleft()
                    self.q.append(dp)


class CacheData(ProxyDataFlow):
    """
    Completely cache the first pass of a DataFlow in memory,
    and produce from the cache thereafter.

    NOTE: The user should not stop the iterator before it has reached the end.
        Otherwise the cache may be incomplete.
    """
    def __init__(self, ds, shuffle=False):
        """
        Args:
            ds (DataFlow): input DataFlow.
            shuffle (bool): whether to shuffle the cache before yielding from it.
        """
        self.shuffle = shuffle
        super(CacheData, self).__init__(ds)

    def reset_state(self):
        super(CacheData, self).reset_state()
        self._guard = DataFlowReentrantGuard()
        if self.shuffle:
            self.rng = get_rng(self)
        self.buffer = []

    def __iter__(self):
        with self._guard:
            if len(self.buffer):
                if self.shuffle:
                    self.rng.shuffle(self.buffer)
                yield from self.buffer
            else:
                for dp in self.ds:
                    yield dp
                    self.buffer.append(dp)


class PrintData(ProxyDataFlow):
    """
    Behave like an identity proxy, but print shape and range of the first few datapoints.
    Good for debugging.

    Example:
        Place it somewhere in your dataflow like

        .. code-block:: python

            def create_my_dataflow():
                ds = SomeDataSource('path/to/lmdb')
                ds = SomeInscrutableMappings(ds)
                ds = PrintData(ds, num=2, max_list=2)
                return ds
            ds = create_my_dataflow()
            # other code that uses ds

        When datapoints are taken from the dataflow, it will print outputs like:

        .. code-block:: none

            [0110 09:22:21 @common.py:589] DataFlow Info:
            datapoint 0<2 with 4 components consists of
               0: float with value 0.0816501893251
               1: ndarray:int32 of shape (64,) in range [0, 10]
               2: ndarray:float32 of shape (64, 64) in range [-1.2248, 1.2177]
               3: list of len 50
                  0: ndarray:int32 of shape (64, 64) in range [-128, 80]
                  1: ndarray:float32 of shape (64, 64) in range [0.8400, 0.6845]
                  ...
            datapoint 1<2 with 4 components consists of
               0: float with value 5.88252075399
               1: ndarray:int32 of shape (64,) in range [0, 10]
               2: ndarray:float32 of shape (64, 64) with range [-0.9011, 0.8491]
               3: list of len 50
                  0: ndarray:int32 of shape (64, 64) in range [-70, 50]
                  1: ndarray:float32 of shape (64, 64) in range [0.7400, 0.3545]
                  ...
    """

    def __init__(self, ds, num=1, name=None, max_depth=3, max_list=3):
        """
        Args:
            ds (DataFlow): input DataFlow.
            num (int): number of dataflow points to print.
            name (str, optional): name to identify this DataFlow.
            max_depth (int, optional): stop output when too deep recursion in sub elements
            max_list (int, optional): stop output when too many sub elements
        """
        super(PrintData, self).__init__(ds)
        self.num = num
        self.name = name
        self.cnt = 0
        self.max_depth = max_depth
        self.max_list = max_list

    def _analyze_input_data(self, entry, k, depth=1, max_depth=3, max_list=3):
        """
        Gather useful debug information from a datapoint.

        Args:
            entry: the datapoint component, either a list or a dict
            k (int): index of this component in current datapoint
            depth (int, optional): recursion depth
            max_depth, max_list: same as in :meth:`__init__`.

        Returns:
            string: debug message
        """

        class _elementInfo(object):
            def __init__(self, el, pos, depth=0, max_list=3):
                self.shape = ""
                self.type = type(el).__name__
                self.dtype = ""
                self.range = ""

                self.sub_elements = []

                self.ident = " " * (depth * 2)
                self.pos = pos

                numpy_scalar_types = list(itertools.chain(*np.sctypes.values()))

                if isinstance(el, (int, float, bool)):
                    self.range = " with value {}".format(el)
                elif type(el) is np.ndarray:
                    self.shape = " of shape {}".format(el.shape)
                    self.dtype = ":{}".format(str(el.dtype))
                    self.range = " in range [{}, {}]".format(el.min(), el.max())
                elif type(el) in numpy_scalar_types:
                    self.range = " with value {}".format(el)
                elif isinstance(el, (list, tuple)):
                    self.shape = " of len {}".format(len(el))

                    if depth < max_depth:
                        for k, subel in enumerate(el):
                            if k < max_list:
                                self.sub_elements.append(_elementInfo(subel, k, depth + 1, max_list))
                            else:
                                self.sub_elements.append(" " * ((depth + 1) * 2) + '...')
                                break
                    else:
                        if len(el) > 0:
                            self.sub_elements.append(" " * ((depth + 1) * 2) + ' ...')

            def __str__(self):
                strings = []
                vals = (self.ident, self.pos, self.type, self.dtype, self.shape, self.range)
                strings.append("{}{}: {}{}{}{}".format(*vals))

                for k, el in enumerate(self.sub_elements):
                    strings.append(str(el))
                return "\n".join(strings)

        return str(_elementInfo(entry, k, depth, max_list))

    def _get_msg(self, dp):
        msg = [colored(u"datapoint %i/%i with %i components consists of" %
               (self.cnt, self.num, len(dp)), "cyan")]
        is_dict = isinstance(dp, Mapping)
        for k, entry in enumerate(dp):
            if is_dict:
                key, value = entry, dp[entry]
            else:
                key, value = k, entry
            msg.append(self._analyze_input_data(value, key, max_depth=self.max_depth, max_list=self.max_list))
        return u'\n'.join(msg)

    def __iter__(self):
        for dp in self.ds:
            # it is important to place this here! otherwise it mixes the output of multiple PrintData
            if self.cnt == 0:
                label = ' (%s)' % self.name if self.name is not None else ""
                logger.info(colored("Contents of DataFlow%s:" % label, 'cyan'))

            if self.cnt < self.num:
                print(self._get_msg(dp))
                self.cnt += 1
            yield dp

    def reset_state(self):
        super(PrintData, self).reset_state()
        self.cnt = 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Any, Callable, Dict, List, Optional, Tuple, Union
from collections import abc
import warnings

from tqdm import tqdm
from loguru import logger

import torch
from torch import nn
from torch.utils.data.dataloader import DataLoader
from torch.utils.data.dataset import Dataset
from torch.utils.data.distributed import DistributedSampler
from torch.utils.data.sampler import RandomSampler, SequentialSampler

from transformers.optimization import Adafactor, AdamW, get_scheduler

# this is used to supress an undesired warning emitted by pytorch versions 1.4.2-1.7.0
try:
    from torch.optim.lr_scheduler import SAVE_STATE_WARNING
except ImportError:
    SAVE_STATE_WARNING = ""


def reissue_pt_warnings(caught_warnings):
    # Reissue warnings that are not the SAVE_STATE_WARNING
    if len(caught_warnings) > 1:
        for w in caught_warnings:
            if w.category != UserWarning or w.message != SAVE_STATE_WARNING:
                warnings.warn(w.message, w.category)


def is_apex_available():
    import importlib.util
    return importlib.util.find_spec("apex") is not None


if is_apex_available():
    from apex import amp


class Trainer:
    def __init__(
        self,
        model: Union[PreTrainedModel, torch.nn.Module] = None,
        args: TrainingArguments = None,
        #  data_collator: Optional[DataCollator] = None,
        train_dataset: Optional[Dataset] = None,
        eval_dataset: Optional[Dataset] = None,
        tokenizer: Optional["PreTrainedTokenizerBase"] = None,
        compute_metrics: Optional[Callable[[EvalPrediction], Dict]] = None,
        optimizers: Tuple[torch.optim.Optimizer,
                          torch.optim.lr_scheduler.LambdaLR] = (None, None),
    ):
        self.model = model
        self.args = args
        #  self.data_collator = data_collator if data_collator is not None else default_collator
        self.train_dataset = train_dataset
        self.eval_dataset = eval_dataset
        self.tokenizer = tokenizer
        self.compute_metrics = compute_metrics
        self.optimizer, self.lr_scheduler = optimizers

        if hasattr(model, "is_parallelizable"
                   ) and model.is_parallelizable and model.model_parallel:
            self.is_model_parallel = True
        else:
            self.is_model_parallel = False

        # postpone switching model to cuda when:
        # 1. MP - since we are trying to fit a much bigger than 1 gpu model
        # 2. fp16-enabled DeepSpeed loads the model in half the size and it doesn't need .to() anyway
        if not (self.is_model_parallel or args.deepspeed):
            model = model.to(args.device)

        # Force n_gpu to 1 to avoid DataParallel as MP will manage the GPUs
        if self.is_model_parallel:
            self.args._n_gpu = 1

        if self.is_master_process():
            os.makedirs(self.args.output_dir, exist_ok=True)

        if args.max_steps > 0:
            logger.info(
                "max_steps is given, it will override any value given in num_train_epochs"
            )
        # Enforce rules on using datasets with no __len__
        if train_dataset is not None and not isinstance(
                train_dataset, abc.Sized) and args.max_steps <= 0:
            raise ValueError(
                "train_dataset does not implement __len__, max_steps has to be specified"
            )
        if eval_dataset is not None and not isinstance(eval_dataset,
                                                       abc.Sized):
            raise ValueError("eval_dataset must implement __len__")

        # Mixed precision setup
        self.use_apex = False
        self.use_amp = False
        self.fp16_backend = None

        from torch.cuda.amp import autocast
        if args.fp16:
            if args.fp16_backend == "auto":
                self.fp16_backend = "amp"
            else:
                self.fp16_backend = args.fp16_backend
            logger.info(f"Using {self.fp16_backend} fp16 backend")

        if args.fp16 and not args.deepspeed:  # deepspeed manages its own fp16
            if self.fp16_backend == "amp":
                self.use_amp = True
                self.scaler = torch.cuda.amp.GradScaler()
            else:
                if not is_apex_available():
                    raise ImportError(
                        "Using FP16 with APEX but APEX is not installed, please refer to https://www.github.com/nvidia/apex."
                    )
                self.use_apex = True

        # Label smoothing
        if self.args.label_smoothing_factor != 0:
            self.label_smoother = LabelSmoother(
                epsilon=self.args.label_smoothing_factor)
        else:
            self.label_smoother = None

    def create_optimizer_and_scheduler(self, num_training_steps: int):
        """
        Setup the optimizer and the learning rate scheduler.

        We provide a reasonable default that works well. If you want to use something else, you can pass a tuple in the
        Trainer's init through :obj:`optimizers`, or subclass and override this method in a subclass.
        """
        if self.optimizer is None:
            no_decay = ["bias", "LayerNorm.weight"]
            optimizer_grouped_parameters = [
                {
                    "params": [
                        p for n, p in self.model.named_parameters()
                        if not any(nd in n for nd in no_decay)
                    ],
                    "weight_decay":
                    self.args.weight_decay,
                },
                {
                    "params": [
                        p for n, p in self.model.named_parameters()
                        if any(nd in n for nd in no_decay)
                    ],
                    "weight_decay":
                    0.0,
                },
            ]
            optimizer_cls = Adafactor if self.args.adafactor else AdamW
            if self.args.adafactor:
                optimizer_cls = Adafactor
                optimizer_kwargs = {
                    "scale_parameter": False,
                    "relative_step": False
                }
            else:
                optimizer_cls = AdamW
                optimizer_kwargs = {
                    "betas": (self.args.adam_beta1, self.args.adam_beta2),
                    "eps": self.args.adam_epsilon,
                }
            optimizer_kwargs["lr"] = self.args.learning_rate
            if self.sharded_dpp:
                self.optimizer = OSS(
                    params=optimizer_grouped_parameters,
                    optim=optimizer_cls,
                    **optimizer_kwargs,
                )
            else:
                self.optimizer = optimizer_cls(optimizer_grouped_parameters,
                                               **optimizer_kwargs)

        if self.lr_scheduler is None:
            self.lr_scheduler = get_scheduler(
                self.args.lr_scheduler_type,
                self.optimizer,
                num_warmup_steps=self.args.warmup_steps,
                num_training_steps=num_training_steps,
            )

    def _load_optimizer_and_scheduler(self, checkpoint):
        """If optimizer and scheduler states exist, load them."""
        if checkpoint is None:
            return

        if os.path.isfile(os.path.join(
                checkpoint, "optimizer.pt")) and os.path.isfile(
                    os.path.join(checkpoint, "scheduler.pt")):
            # Load in optimizer and scheduler states
            self.optimizer.load_state_dict(
                torch.load(os.path.join(checkpoint, "optimizer.pt"),
                           map_location=self.args.device))
            with warnings.catch_warnings(record=True) as caught_warnings:
                self.lr_scheduler.load_state_dict(
                    torch.load(os.path.join(checkpoint, "scheduler.pt")))
            reissue_pt_warnings(caught_warnings)

        if self.deepspeed:
            # Not sure how to check if there is a saved deepspeed checkpoint, but since it just return None if it fails to find a deepspeed checkpoint this is sort of a check-n-load function
            self.deepspeed.load_checkpoint(checkpoint,
                                           load_optimizer_states=True,
                                           load_lr_scheduler_states=True)

    def is_master_process(self) -> bool:
        return self.args.local_rank == -1 or torch.distributed.get_rank() == 0

    def _get_train_sampler(self) -> Optional[torch.utils.data.sampler.Sampler]:
        if isinstance(self.train_dataset,
                      torch.utils.data.IterableDataset) or not isinstance(
                          self.train_dataset, collections.abc.Sized):
            return None

        # Gather the number of processes and this process index.
        if (self.args.parallel_mode == ParallelMode.DISTRIBUTED or
                self.args.parallel_mode == ParallelMode.SAGEMAKER_DISTRIBUTED):
            num_processes = torch.distributed.get_world_size()
            process_index = torch.distributed.get_rank()
        else:
            num_processes = 1
            process_index = 0

        # Build the sampler.
        if num_processes <= 1:
            return RandomSampler(self.train_dataset)
        else:
            return DistributedSampler(self.train_dataset,
                                      num_replicas=num_processes,
                                      rank=process_index)

    def get_train_dataloader(self) -> DataLoader:
        if self.train_dataset is None:
            raise ValueError("Trainer: training requires a train_dataset.")
        train_sampler = self._get_train_sampler()

        return DataLoader(
            self.train_dataset,
            batch_size=self.args.train_batch_size,
            sampler=train_sampler,
            collate_fn=self.data_collator,
            drop_last=self.args.dataloader_drop_last,
            num_workers=self.args.dataloader_num_workers,
            pin_memory=self.args.dataloader_pin_memory,
        )

    def train(
        self,
        resume_from_checkpoint: Optional[str] = None,
        **kwargs,
    ):
        if "model_path" in kwargs:
            resume_from_checkpoint = kwargs.pop("model_path")
            warnings.warn(
                "`model_path` is deprecated and will be removed in a future version. Use `resume_from_checkpoint` "
                "instead.",
                FutureWarning,
            )

        model_reloaded = False
        # Load potential model checkpoint
        PYTORCH_WEIGHTS_NAME = "pytorch_model.bin"
        if resume_from_checkpoint is not None and os.path.isfile(
                os.path.join(resume_from_checkpoint, PYTORCH_WEIGHTS_NAME)):
            logger.info(f"Loading model from {resume_from_checkpoint}).")
            if isinstance(self.model, PreTrainedModel):
                self.model = self.model.from_pretrained(resume_from_checkpoint)
                model_reloaded = True
            else:
                state_dict = torch.load(
                    os.path.join(resume_from_checkpoint, PYTORCH_WEIGHTS_NAME))
                self.model.load_state_dict(state_dict)

        # If model was re-initialized, put it on the right device and update self.model_wrapped
        if model_reloaded:
            if not self.is_model_parallel:
                self.model = self.model.to(self.args.device)
            #  self.model_wrapped = self.model

        train_dataloader = self.get_train_dataloader()

        train_dataset_is_sized = isinstance(self.train_dataset,
                                            collections.abc.Sized)
        if train_dataset_is_sized:
            num_update_steps_per_epoch = len(
                train_dataloader) // self.args.gradient_accumulation_steps
            num_update_steps_per_epoch = max(num_update_steps_per_epoch, 1)
            if self.args.max_steps > 0:
                max_steps = self.args.max_steps
                num_train_epochs = self.args.max_steps // num_update_steps_per_epoch + int(
                    self.args.max_steps % num_update_steps_per_epoch > 0)
            else:
                max_steps = math.ceil(self.args.num_train_epochs *
                                      num_update_steps_per_epoch)
                num_train_epochs = math.ceil(self.args.num_train_epochs)
        else:
            # see __init__. max_steps is set when the dataset has no __len__
            max_steps = self.args.max_steps
            num_train_epochs = 1
            num_update_steps_per_epoch = max_steps

        self.create_optimizer_and_scheduler(num_training_steps=max_steps)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import random, copy, json
from tqdm import tqdm
from loguru import logger
from theta.utils import seg_generator
from dataclasses import dataclass, field
import numpy as np
#  import mlflow

from ..utils import seg_generator


#  @dataclass(frozen=False)
class InputExample:
    """
    A single training/test example for simple sequence classification.

    Args:
        guid: Unique id for the example.
        text_a: string. The untokenized text of the first sequence. For single
            sequence tasks, only this sequence must be specified.
        text_b: (Optional) string. The untokenized text of the second sequence.
            Only must be specified for sequence pair tasks.
        label: (Optional) string. The label of the example. This should be
            specified for train and dev examples, but not for test examples.
    """

    #  guid: str
    #  text_a: str
    #  text_b: Optional[str] = None
    #  label: Optional[str] = None
    def __init__(self, guid, text_a, text_b=None, label=None):
        self.guid = guid
        self.text_a = text_a
        self.text_b = text_b
        self.label = label

    def to_json_string(self):
        """Serializes this instance to a JSON string."""
        return json.dumps(dataclasses.asdict(self), indent=2,
                          sort_keys=True) + "\n"


def load_glue_examples(data_generator, examples_file):

    examples = []

    for guid, text_a, text_b, label in data_generator(examples_file):
        examples.append(
            InputExample(guid=guid, text_a=text_a, text_b=text_b, label=label))
    logger.info(f"Loaded {len(examples)} examples.")

    return examples


def load_train_val_examples(args,
                            train_data_generator,
                            glue_labels,
                            shuffle=True,
                            train_rate=0.9,
                            num_augments=0):
    all_train_examples = load_glue_examples(train_data_generator,
                                            args.train_file)

    if args.train_sample_rate < 1.0:
        num_samples = int(len(all_train_examples) * args.train_sample_rate)
        all_train_examples = all_train_examples[:num_samples]

    # 切分训练集和验证集
    # theta 提供split_train_eval_examples辅助函数
    from theta.utils import split_train_eval_examples
    train_examples, val_examples = split_train_eval_examples(
        all_train_examples,
        train_rate=args.train_rate,
        fold=args.fold,
        shuffle=shuffle)

    #  random.shuffle(all_train_examples)
    #  num_train_examples = int(len(all_train_examples) * args.train_rate)
    #  val_examples = all_train_examples[num_train_examples:]
    #  train_examples = all_train_examples[:num_train_examples]

    logger.info(f"Loaded {len(train_examples)} train examples, "
                f"{len(val_examples)} val examples.")
    return train_examples, val_examples


def load_test_examples(args, test_data_generator):
    test_base_examples = load_glue_examples(test_data_generator,
                                            args.test_file)

    logger.info(f"Loaded {len(test_base_examples)} test examples.")
    return test_base_examples


def show_glue_datainfo(glue_labels, train_data_generator, train_file,
                       test_data_generator, test_file):
    train_lengths = [
        len(text_a)
        for guid, text_a, _, label in train_data_generator(train_file)
    ]

    all_labels = []
    for _, _, _, label in train_data_generator(train_file):
        all_labels.append(label)
        if label not in glue_labels:
            glue_labels.append(label)

    test_lengths = [
        len(text_a)
        for guid, text_a, _, label in test_data_generator(test_file)
    ]

    logger.info(f"****** glue_labels ******")
    logger.info(f"{glue_labels}")
    if isinstance(all_labels[0], str):
        from collections import Counter
        logger.info(f"{Counter(all_labels).most_common()}")
    logger.info(f"")

    logger.info(f"Train samples: {len(train_lengths)}")
    logger.info(f"Test samples: {len(test_lengths)}")

    logger.info(f"")
    import numpy as np
    logger.info(f"****** train lengths ******")
    logger.info(f"mean: {np.mean(train_lengths):.2f}")
    logger.info(f"std: {np.std(train_lengths):.2f}")
    logger.info(f"min: {np.min(train_lengths)}")
    logger.info(f"max: {np.max(train_lengths)}")
    logger.info(f"")

    logger.info(f"****** test lengths ******")
    logger.info(f"mean: {np.mean(test_lengths):.2f}")
    logger.info(f"std: {np.std(test_lengths):.2f}")
    logger.info(f"min: {np.min(test_lengths)}")
    logger.info(f"max: {np.max(test_lengths)}")
    logger.info(f"")


def save_glue_preds(args, preds, test_examples, probs=None):
    assert len(test_examples) == len(preds)
    reviews_file = args.reviews_file

    reviews = {}
    for i, (input_example, v) in enumerate(zip(test_examples, preds)):
        guid = input_example.guid
        text_a = input_example.text_a or ""
        text_b = input_example.text_b or ""
        if isinstance(v, list) or isinstance(v, np.ndarray):
            label = [args.id2label[i] for i, x in enumerate(v) if x]
        else:
            label = args.id2label[v]

        prob = None
        if isinstance(probs, np.ndarray):
            prob = [f"{x:.3f}" for x in list(probs[i])]
        elif isinstance(probs, list):
            prob = [f"{x:.3f}" for x in probs[i]]
        else:
            prob = None

        reviews[guid] = {
            'guid': guid,
            'text_a': text_a,
            'text_b': text_b,
            'label': label,
            'probs': prob
        }

    json.dump(reviews, open(reviews_file, 'w'), ensure_ascii=False, indent=2)
    #  with open(reviews_file, 'w') as fw:
    #      fw.write("guid\ttext_a\ttext_b\tlabel\n")
    #      for input_example, v in zip(test_examples, preds):
    #          guid = input_example.guid
    #          text_a = input_example.text_a or ""
    #          text_b = input_example.text_b or ""
    #          label = args.id2label[v]
    #          fw.write(f"{guid}\t{text_a}\t{text_b}\t{label}\n")

    logger.info(f"Total {len(preds)} lines saved to {reviews_file}")

    #  ----- Tracking -----
    #  if args.do_experiment:
    #      logger.debug(f"mlflow tracking uri: {mlflow.get_tracking_uri()}")
    #      mlflow.log_param(f"{args.dataset_name}_reviews_file", reviews_file)
    #      mlflow.log_artifact(reviews_file, args.artifact_path)

    return reviews_file



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import torch
from tqdm import tqdm
from loguru import logger
import numpy as np
from collections import defaultdict


def crf_decode(decode_tokens, raw_text, id2ent):
    """
    CRF 解码，用于解码 time loc 的提取
    """
    predict_entities = {}

    decode_tokens = decode_tokens[1:-1]  # 除去 CLS SEP token

    index_ = 0

    while index_ < len(decode_tokens):

        token_label = id2ent[decode_tokens[index_]].split('-')

        if token_label[0].startswith('S'):
            token_type = token_label[1]
            tmp_ent = raw_text[index_]

            if token_type not in predict_entities:
                predict_entities[token_type] = [(tmp_ent, index_)]
            else:
                predict_entities[token_type].append((tmp_ent, int(index_)))

            index_ += 1

        elif token_label[0].startswith('B'):
            token_type = token_label[1]
            start_index = index_

            index_ += 1
            while index_ < len(decode_tokens):
                temp_token_label = id2ent[decode_tokens[index_]].split('-')

                if temp_token_label[0].startswith(
                        'I') and token_type == temp_token_label[1]:
                    index_ += 1
                elif temp_token_label[0].startswith(
                        'E') and token_type == temp_token_label[1]:
                    end_index = index_
                    index_ += 1

                    tmp_ent = raw_text[start_index:end_index + 1]

                    if token_type not in predict_entities:
                        predict_entities[token_type] = [(tmp_ent, start_index)]
                    else:
                        predict_entities[token_type].append(
                            (tmp_ent, int(start_index)))

                    break
                else:
                    break
        else:
            index_ += 1

    return predict_entities


# 严格解码 baseline
def span_decode(start_logits, end_logits, raw_text, id2ent):
    predict_entities = defaultdict(list)

    start_pred = np.argmax(start_logits, -1)
    end_pred = np.argmax(end_logits, -1)

    for i, s_type in enumerate(start_pred):
        if s_type == 0:
            continue
        for j, e_type in enumerate(end_pred[i:]):
            if s_type == e_type:
                tmp_ent = raw_text[i:i + j + 1]
                predict_entities[id2ent[s_type]].append((tmp_ent, i))
                break

    return predict_entities


# 严格解码 baseline
def mrc_decode(start_logits, end_logits, raw_text):
    predict_entities = []
    start_pred = np.argmax(start_logits, -1)
    end_pred = np.argmax(end_logits, -1)

    for i, s_type in enumerate(start_pred):
        if s_type == 0:
            continue
        for j, e_type in enumerate(end_pred[i:]):
            if s_type == e_type:
                tmp_ent = raw_text[i:i + j + 1]
                predict_entities.append((tmp_ent, i))
                break

    return predict_entities



# -*- coding: utf-8 -*-
# File: misc.py

import cv2

from ...utils import logger
from ...utils.argtools import shape2d
from .base import ImageAugmentor
from .transform import ResizeTransform, NoOpTransform, FlipTransform, TransposeTransform

__all__ = ['Flip', 'Resize', 'RandomResize', 'ResizeShortestEdge', 'Transpose']


class Flip(ImageAugmentor):
    """
    Random flip the image either horizontally or vertically.
    """
    def __init__(self, horiz=False, vert=False, prob=0.5):
        """
        Args:
            horiz (bool): use horizontal flip.
            vert (bool): use vertical flip.
            prob (float): probability of flip.
        """
        super(Flip, self).__init__()
        if horiz and vert:
            raise ValueError("Cannot do both horiz and vert. Please use two Flip instead.")
        if not horiz and not vert:
            raise ValueError("At least one of horiz or vert has to be True!")
        self._init(locals())

    def get_transform(self, img):
        h, w = img.shape[:2]
        do = self._rand_range() < self.prob
        if not do:
            return NoOpTransform()
        else:
            return FlipTransform(h, w, self.horiz)


class Resize(ImageAugmentor):
    """ Resize image to a target size"""

    def __init__(self, shape, interp=cv2.INTER_LINEAR):
        """
        Args:
            shape: (h, w) tuple or a int
            interp: cv2 interpolation method
        """
        shape = tuple(shape2d(shape))
        self._init(locals())

    def get_transform(self, img):
        return ResizeTransform(
            img.shape[0], img.shape[1],
            self.shape[0], self.shape[1], self.interp)


class ResizeShortestEdge(ImageAugmentor):
    """
    Resize the shortest edge to a certain number while
    keeping the aspect ratio.
    """

    def __init__(self, size, interp=cv2.INTER_LINEAR):
        """
        Args:
            size (int): the size to resize the shortest edge to.
        """
        size = int(size)
        self._init(locals())

    def get_transform(self, img):
        h, w = img.shape[:2]
        scale = self.size * 1.0 / min(h, w)
        if h < w:
            newh, neww = self.size, int(scale * w + 0.5)
        else:
            newh, neww = int(scale * h + 0.5), self.size
        return ResizeTransform(h, w, newh, neww, self.interp)


class RandomResize(ImageAugmentor):
    """ Randomly rescale width and height of the image."""

    def __init__(self, xrange, yrange=None, minimum=(0, 0), aspect_ratio_thres=0.15,
                 interp=cv2.INTER_LINEAR):
        """
        Args:
            xrange (tuple): a (min, max) tuple. If is floating point, the
                tuple defines the range of scaling ratio of new width, e.g. (0.9, 1.2).
                If is integer, the tuple defines the range of new width in pixels, e.g. (200, 350).
            yrange (tuple): similar to xrange, but for height. Should be None when aspect_ratio_thres==0.
            minimum (tuple): (xmin, ymin) in pixels. To avoid scaling down too much.
            aspect_ratio_thres (float): discard samples which change aspect ratio
                larger than this threshold. Set to 0 to keep aspect ratio.
            interp: cv2 interpolation method
        """
        super(RandomResize, self).__init__()
        assert aspect_ratio_thres >= 0
        self._init(locals())

        def is_float(tp):
            return isinstance(tp[0], float) or isinstance(tp[1], float)

        if yrange is not None:
            assert is_float(xrange) == is_float(yrange), "xrange and yrange has different type!"
        self._is_scale = is_float(xrange)

        if aspect_ratio_thres == 0:
            if self._is_scale:
                assert xrange == yrange or yrange is None
            else:
                if yrange is not None:
                    logger.warn("aspect_ratio_thres==0, yrange is not used!")

    def get_transform(self, img):
        cnt = 0
        h, w = img.shape[:2]

        def get_dest_size():
            if self._is_scale:
                sx = self._rand_range(*self.xrange)
                if self.aspect_ratio_thres == 0:
                    sy = sx
                else:
                    sy = self._rand_range(*self.yrange)
                destX = max(sx * w, self.minimum[0])
                destY = max(sy * h, self.minimum[1])
            else:
                sx = self._rand_range(*self.xrange)
                if self.aspect_ratio_thres == 0:
                    sy = sx * 1.0 / w * h
                else:
                    sy = self._rand_range(*self.yrange)
                destX = max(sx, self.minimum[0])
                destY = max(sy, self.minimum[1])
            return (int(destX + 0.5), int(destY + 0.5))

        while True:
            destX, destY = get_dest_size()
            if self.aspect_ratio_thres > 0:  # don't check when thres == 0
                oldr = w * 1.0 / h
                newr = destX * 1.0 / destY
                diff = abs(newr - oldr) / oldr
                if diff >= self.aspect_ratio_thres + 1e-5:
                    cnt += 1
                    if cnt > 50:
                        logger.warn("RandomResize failed to augment an image")
                        return ResizeTransform(h, w, h, w, self.interp)
                    continue
            return ResizeTransform(h, w, destY, destX, self.interp)


class Transpose(ImageAugmentor):
    """
    Random transpose the image
    """
    def __init__(self, prob=0.5):
        """
        Args:
            prob (float): probability of transpose.
        """
        super(Transpose, self).__init__()
        self.prob = prob

    def get_transform(self, _):
        if self.rng.rand() < self.prob:
            return TransposeTransform()
        else:
            return NoOpTransform()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from tqdm import tqdm
from loguru import logger

from datasets.arrow_dataset import Dataset as ArrowDataset
from datasets.dataset_dict import DatasetDict


def yield_data_generator(iterable):
    def _generator():
        for x in tqdm(iterable):
            yield x

    return _generator


def check_labels(special_labels, labels_list):
    unknown_labels = set(labels_list) - set(special_labels)
    if any(unknown_labels):
        logger.warning(
            f"Labels {unknown_labels} not in special_labels: {special_labels}")
    notexist_labels = set(special_labels) - set(labels_list)
    if any(notexist_labels):
        logger.warning(
            f"Labels {notexist_labels} not in labels_list: {set(labels_list)}")


def convert_glue_samples_to_hf_dataset(samples, glue_labels, is_test=False):
    idx_list = []
    text_a_list = []
    text_b_list = []
    labels_list = []

    for guid, text_a, text_b, label in tqdm(samples):
        idx_list.append(guid)
        text_a_list.append(text_a)
        text_b_list.append(text_b)
        if not is_test:
            labels_list.append(label)

    data_dict = {
        'idx': idx_list,
        'sentence1': text_a_list,
    }
    if any(text_b_list):
        data_dict['sentence2'] = text_b_list
    if not is_test:
        check_labels(glue_labels, labels_list)
        data_dict['label'] = labels_list

    dataset = ArrowDataset.from_dict(data_dict)

    return dataset


def encode_glue_samples_to_hf_dataset(samples,
                                      encode_datasets_fn,
                                      glue_labels,
                                      tokenizer,
                                      data_args,
                                      is_test=False):
    arrow_dataset = convert_glue_samples_to_hf_dataset(samples,
                                                       glue_labels,
                                                       is_test=is_test)

    hf_datasets = DatasetDict({
        'dataset': arrow_dataset,
    })

    label_to_id = {v: i for i, v in enumerate(glue_labels)}
    sentence1_key = "sentence1"
    sentence2_key = "sentence2"

    hf_datasets = encode_datasets_fn(tokenizer, hf_datasets, data_args,
                                     label_to_id, sentence1_key, sentence2_key)

    return hf_datasets['dataset']



# -*- coding: utf-8 -*-
# File: logger.py

"""
The logger module itself has the common logging functions of Python's
:class:`logging.Logger`. For example:

.. code-block:: python

    from tensorpack.utils import logger
    logger.set_logger_dir('train_log/test')
    logger.info("Test")
    logger.error("Error happened!")
"""


import logging
import os
import os.path
import shutil
import sys
from datetime import datetime
from six.moves import input
from termcolor import colored

__all__ = ['set_logger_dir', 'auto_set_dir', 'get_logger_dir']


class _MyFormatter(logging.Formatter):
    def format(self, record):
        date = colored('[%(asctime)s @%(filename)s:%(lineno)d]', 'green')
        msg = '%(message)s'
        if record.levelno == logging.WARNING:
            fmt = date + ' ' + colored('WRN', 'red', attrs=['blink']) + ' ' + msg
        elif record.levelno == logging.ERROR or record.levelno == logging.CRITICAL:
            fmt = date + ' ' + colored('ERR', 'red', attrs=['blink', 'underline']) + ' ' + msg
        elif record.levelno == logging.DEBUG:
            fmt = date + ' ' + colored('DBG', 'yellow', attrs=['blink']) + ' ' + msg
        else:
            fmt = date + ' ' + msg
        if hasattr(self, '_style'):
            # Python3 compatibility
            self._style._fmt = fmt
        self._fmt = fmt
        return super(_MyFormatter, self).format(record)


def _getlogger():
    # this file is synced to "dataflow" package as well
    package_name = "dataflow" if __name__.startswith("dataflow") else "tensorpack"
    logger = logging.getLogger(package_name)
    logger.propagate = False
    logger.setLevel(logging.INFO)
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(_MyFormatter(datefmt='%m%d %H:%M:%S'))
    logger.addHandler(handler)
    return logger


_logger = _getlogger()
_LOGGING_METHOD = ['info', 'warning', 'error', 'critical', 'exception', 'debug', 'setLevel', 'addFilter']
# export logger functions
for func in _LOGGING_METHOD:
    locals()[func] = getattr(_logger, func)
    __all__.append(func)
# 'warn' is deprecated in logging module
warn = _logger.warning
__all__.append('warn')


def _get_time_str():
    return datetime.now().strftime('%m%d-%H%M%S')


# globals: logger file and directory:
LOG_DIR = None
_FILE_HANDLER = None


def _set_file(path):
    global _FILE_HANDLER
    if os.path.isfile(path):
        backup_name = path + '.' + _get_time_str()
        shutil.move(path, backup_name)
        _logger.info("Existing log file '{}' backuped to '{}'".format(path, backup_name))  # noqa: F821
    hdl = logging.FileHandler(
        filename=path, encoding='utf-8', mode='w')
    hdl.setFormatter(_MyFormatter(datefmt='%m%d %H:%M:%S'))

    _FILE_HANDLER = hdl
    _logger.addHandler(hdl)
    _logger.info("Argv: " + ' '.join(sys.argv))


def set_logger_dir(dirname, action=None):
    """
    Set the directory for global logging.

    Args:
        dirname(str): log directory
        action(str): an action of ["k","d","q"] to be performed
            when the directory exists. Will ask user by default.

                "d": delete the directory. Note that the deletion may fail when
                the directory is used by tensorboard.

                "k": keep the directory. This is useful when you resume from a
                previous training and want the directory to look as if the
                training was not interrupted.
                Note that this option does not load old models or any other
                old states for you. It simply does nothing.

    """
    dirname = os.path.normpath(dirname)
    global LOG_DIR, _FILE_HANDLER
    if _FILE_HANDLER:
        # unload and close the old file handler, so that we may safely delete the logger directory
        _logger.removeHandler(_FILE_HANDLER)
        del _FILE_HANDLER

    def dir_nonempty(dirname):
        # If directory exists and nonempty (ignore hidden files), prompt for action
        return os.path.isdir(dirname) and len([x for x in os.listdir(dirname) if x[0] != '.'])

    if dir_nonempty(dirname):
        if not action:
            _logger.warning("""\
Log directory {} exists! Use 'd' to delete it. """.format(dirname))
            _logger.warning("""\
If you're resuming from a previous run, you can choose to keep it.
Press any other key to exit. """)
        while not action:
            action = input("Select Action: k (keep) / d (delete) / q (quit):").lower().strip()
        act = action
        if act == 'b':
            backup_name = dirname + _get_time_str()
            shutil.move(dirname, backup_name)
            info("Directory '{}' backuped to '{}'".format(dirname, backup_name))  # noqa: F821
        elif act == 'd':
            shutil.rmtree(dirname, ignore_errors=True)
            if dir_nonempty(dirname):
                shutil.rmtree(dirname, ignore_errors=False)
        elif act == 'n':
            dirname = dirname + _get_time_str()
            info("Use a new log directory {}".format(dirname))  # noqa: F821
        elif act == 'k':
            pass
        else:
            raise OSError("Directory {} exits!".format(dirname))
    LOG_DIR = dirname
    from .fs import mkdir_p
    mkdir_p(dirname)
    _set_file(os.path.join(dirname, 'log.log'))


def auto_set_dir(action=None, name=None):
    """
    Use :func:`logger.set_logger_dir` to set log directory to
    "./train_log/{scriptname}:{name}". "scriptname" is the name of the main python file currently running"""
    mod = sys.modules['__main__']
    basename = os.path.basename(mod.__file__)
    auto_dirname = os.path.join('train_log', basename[:basename.rfind('.')])
    if name:
        auto_dirname += '_%s' % name if os.name == 'nt' else ':%s' % name
    set_logger_dir(auto_dirname, action=action)


def get_logger_dir():
    """
    Returns:
        The logger directory, or None if not set.
        The directory is used for general logging, tensorboard events, checkpoints, etc.
    """
    return LOG_DIR



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import copy
import torch
import logging
from torch.cuda.amp import autocast as ac
from torch.utils.data import DataLoader, RandomSampler
from transformers import AdamW, get_linear_schedule_with_warmup
from attack_utils import FGM, PGD

logger = logging.getLogger(__name__)


def save_model(opt, model, global_step):
    output_dir = os.path.join(opt.output_dir,
                              'checkpoint-{}'.format(global_step))
    if not os.path.exists(output_dir):
        os.makedirs(output_dir, exist_ok=True)

    # take care of model distributed / parallel training
    model_to_save = (model.module if hasattr(model, "module") else model)
    logger.info(
        f'Saving model & optimizer & scheduler checkpoint to {output_dir}')
    torch.save(model_to_save.state_dict(),
               os.path.join(output_dir, 'model.pt'))


def build_optimizer_and_scheduler(opt, model, t_total):
    module = (model.module if hasattr(model, "module") else model)

    # 差分学习率
    no_decay = ["bias", "LayerNorm.weight"]
    model_param = list(module.named_parameters())

    bert_param_optimizer = []
    other_param_optimizer = []

    for name, para in model_param:
        space = name.split('.')
        if space[0] == 'bert_module':
            bert_param_optimizer.append((name, para))
        else:
            other_param_optimizer.append((name, para))

    optimizer_grouped_parameters = [
        # bert other module
        {
            "params": [
                p for n, p in bert_param_optimizer
                if not any(nd in n for nd in no_decay)
            ],
            "weight_decay":
            opt.weight_decay,
            'lr':
            opt.lr
        },
        {
            "params": [
                p for n, p in bert_param_optimizer
                if any(nd in n for nd in no_decay)
            ],
            "weight_decay":
            0.0,
            'lr':
            opt.lr
        },

        # 其他模块，差分学习率
        {
            "params": [
                p for n, p in other_param_optimizer
                if not any(nd in n for nd in no_decay)
            ],
            "weight_decay":
            opt.weight_decay,
            'lr':
            opt.other_lr
        },
        {
            "params": [
                p for n, p in other_param_optimizer
                if any(nd in n for nd in no_decay)
            ],
            "weight_decay":
            0.0,
            'lr':
            opt.other_lr
        },
    ]

    optimizer = AdamW(optimizer_grouped_parameters,
                      lr=opt.lr,
                      eps=opt.adam_epsilon)
    scheduler = get_linear_schedule_with_warmup(
        optimizer,
        num_warmup_steps=int(opt.warmup_proportion * t_total),
        num_training_steps=t_total)

    return optimizer, scheduler


def load_model_and_parallel(model, gpu_ids, ckpt_path=None, strict=True):
    """
    加载模型 & 放置到 GPU 中（单卡 / 多卡）
    """
    gpu_ids = gpu_ids.split(',')

    # set to device to the first cuda
    device = torch.device("cpu" if gpu_ids[0] == '-1' else "cuda:" +
                          gpu_ids[0])

    if ckpt_path is not None:
        logger.info(f'Load ckpt from {ckpt_path}')
        model.load_state_dict(torch.load(ckpt_path,
                                         map_location=torch.device('cpu')),
                              strict=strict)

    model.to(device)

    if len(gpu_ids) > 1:
        logger.info(f'Use multi gpus in: {gpu_ids}')
        gpu_ids = [int(x) for x in gpu_ids]
        model = torch.nn.DataParallel(model, device_ids=gpu_ids)
    else:
        logger.info(f'Use single gpu in: {gpu_ids}')

    return model, device


def get_model_path_list(base_dir):
    """
    从文件夹中获取 model.pt 的路径
    """
    model_lists = []

    for root, dirs, files in os.walk(base_dir):
        for _file in files:
            if 'model.pt' == _file:
                model_lists.append(os.path.join(root, _file))

    model_lists = sorted(
        model_lists,
        key=lambda x: (x.split('/')[-3], int(x.split('/')[-2].split('-')[-1])))

    return model_lists


def swa(model, model_dir, swa_start=1):
    """
    swa 滑动平均模型，一般在训练平稳阶段再使用 SWA
    """
    model_path_list = get_model_path_list(model_dir)

    assert 1 <= swa_start < len(model_path_list) - 1, \
        f'Using swa, swa start should smaller than {len(model_path_list) - 1} and bigger than 0'

    swa_model = copy.deepcopy(model)
    swa_n = 0.

    with torch.no_grad():
        for _ckpt in model_path_list[swa_start:]:
            logger.info(f'Load model from {_ckpt}')
            model.load_state_dict(
                torch.load(_ckpt, map_location=torch.device('cpu')))
            tmp_para_dict = dict(model.named_parameters())

            alpha = 1. / (swa_n + 1.)

            for name, para in swa_model.named_parameters():
                para.copy_(tmp_para_dict[name].data.clone() * alpha +
                           para.data.clone() * (1. - alpha))

            swa_n += 1

    # use 100000 to represent swa to avoid clash
    swa_model_dir = os.path.join(model_dir, f'checkpoint-100000')
    if not os.path.exists(swa_model_dir):
        os.mkdir(swa_model_dir)

    logger.info(f'Save swa model in: {swa_model_dir}')

    swa_model_path = os.path.join(swa_model_dir, 'model.pt')

    torch.save(swa_model.state_dict(), swa_model_path)

    return swa_model


def train(opt, model, train_dataset):
    swa_raw_model = copy.deepcopy(model)

    train_sampler = RandomSampler(train_dataset)

    train_loader = DataLoader(dataset=train_dataset,
                              batch_size=opt.train_batch_size,
                              sampler=train_sampler,
                              num_workers=0)

    scaler = None
    if opt.use_fp16:
        scaler = torch.cuda.amp.GradScaler()

    model, device = load_model_and_parallel(model, opt.gpu_ids)

    use_n_gpus = False
    if hasattr(model, "module"):
        use_n_gpus = True

    t_total = len(train_loader) * opt.train_epochs

    optimizer, scheduler = build_optimizer_and_scheduler(opt, model, t_total)

    # Train
    logger.info("***** Running training *****")
    logger.info(f"  Num Examples = {len(train_dataset)}")
    logger.info(f"  Num Epochs = {opt.train_epochs}")
    logger.info(f"  Total training batch size = {opt.train_batch_size}")
    logger.info(f"  Total optimization steps = {t_total}")

    global_step = 0

    model.zero_grad()

    fgm, pgd = None, None

    attack_train_mode = opt.attack_train.lower()
    if attack_train_mode == 'fgm':
        fgm = FGM(model=model)
    elif attack_train_mode == 'pgd':
        pgd = PGD(model=model)

    pgd_k = 3

    save_steps = t_total // opt.train_epochs
    eval_steps = save_steps

    logger.info(
        f'Save model in {save_steps} steps; Eval model in {eval_steps} steps')

    log_loss_steps = 20

    avg_loss = 0.

    for epoch in range(opt.train_epochs):

        for step, batch_data in enumerate(train_loader):

            model.train()

            for key in batch_data.keys():
                batch_data[key] = batch_data[key].to(device)

            if opt.use_fp16:
                with ac():
                    loss = model(**batch_data)[0]
            else:
                loss = model(**batch_data)[0]

            if use_n_gpus:
                loss = loss.mean()

            if opt.use_fp16:
                scaler.scale(loss).backward()
            else:
                loss.backward()

            if fgm is not None:
                fgm.attack()

                if opt.use_fp16:
                    with ac():
                        loss_adv = model(**batch_data)[0]
                else:
                    loss_adv = model(**batch_data)[0]

                if use_n_gpus:
                    loss_adv = loss_adv.mean()

                if opt.use_fp16:
                    scaler.scale(loss_adv).backward()
                else:
                    loss_adv.backward()

                fgm.restore()

            elif pgd is not None:
                pgd.backup_grad()

                for _t in range(pgd_k):
                    pgd.attack(is_first_attack=(_t == 0))

                    if _t != pgd_k - 1:
                        model.zero_grad()
                    else:
                        pgd.restore_grad()

                    if opt.use_fp16:
                        with ac():
                            loss_adv = model(**batch_data)[0]
                    else:
                        loss_adv = model(**batch_data)[0]

                    if use_n_gpus:
                        loss_adv = loss_adv.mean()

                    if opt.use_fp16:
                        scaler.scale(loss_adv).backward()
                    else:
                        loss_adv.backward()

                pgd.restore()

            if opt.use_fp16:
                scaler.unscale_(optimizer)

            torch.nn.utils.clip_grad_norm_(model.parameters(),
                                           opt.max_grad_norm)

            # optimizer.step()
            if opt.use_fp16:
                scaler.step(optimizer)
                scaler.update()
            else:
                optimizer.step()

            scheduler.step()
            model.zero_grad()

            global_step += 1

            if global_step % log_loss_steps == 0:
                avg_loss /= log_loss_steps
                logger.info('Step: %d / %d ----> total loss: %.5f' %
                            (global_step, t_total, avg_loss))
                avg_loss = 0.
            else:
                avg_loss += loss.item()

            if global_step % save_steps == 0:
                save_model(opt, model, global_step)

    swa(swa_raw_model, opt.output_dir, swa_start=opt.swa_start)

    # clear cuda cache to avoid OOM
    torch.cuda.empty_cache()
    logger.info('Train done')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Union
from tqdm import tqdm
from loguru import logger

from theta.nlp.data.samples import GlueSamples
from theta.nlp.tasks import GlueData, GlueModel, GlueTask


class TextClassificationPipeline:
    def __init__(self, checkpoint_path, num_labels, glue_labels):
        self.checkpoint_path = checkpoint_path
        self.num_labels = num_labels
        self.glue_labels = glue_labels

    def __call__(self,
                 text: Union[str, list],
                 max_length=512,
                 glue_labels=None):

        # -------------------- Data --------------------
        def data_generator():
            if isinstance(text, str):
                yield 0, text, None, None
            elif isinstance(text, list):
                texts = text
                for i, x in enumerate(texts):
                    yield f"{i}", x, None, None
            else:
                raise TypeError(
                    f"text({type(text)}) type must be str or list.")

        test_samples = GlueSamples(labels_list=glue_labels if glue_labels else
                                   [f"{x}" for x in range(self.num_labels)],
                                   data_generator=data_generator)
        data_args = {
            'max_length': max_length,
        }
        task_data = GlueData(data_args,
                             train_samples=None,
                             test_samples=test_samples)

        # -------------------- model --------------------
        task_model = GlueModel(task_args, self.num_labels)



# -*- coding: utf-8 -*-
# File: image.py


import copy as copy_mod
import numpy as np
from contextlib import contextmanager

from ..utils import logger
from ..utils.argtools import shape2d
from .base import RNGDataFlow
from .common import MapData, MapDataComponent

__all__ = ['ImageFromFile', 'AugmentImageComponent', 'AugmentImageCoordinates', 'AugmentImageComponents']


def check_dtype(img):
    assert isinstance(img, np.ndarray), "[Augmentor] Needs an numpy array, but got a {}!".format(type(img))
    assert not isinstance(img.dtype, np.integer) or (img.dtype == np.uint8), \
        "[Augmentor] Got image of type {}, use uint8 or floating points instead!".format(img.dtype)


def validate_coords(coords):
    assert coords.ndim == 2, coords.ndim
    assert coords.shape[1] == 2, coords.shape
    assert np.issubdtype(coords.dtype, np.float), coords.dtype


class ExceptionHandler:
    def __init__(self, catch_exceptions=False):
        self._nr_error = 0
        self.catch_exceptions = catch_exceptions

    @contextmanager
    def catch(self):
        try:
            yield
        except Exception:
            self._nr_error += 1
            if not self.catch_exceptions:
                raise
            else:
                if self._nr_error % 100 == 0 or self._nr_error < 10:
                    logger.exception("Got {} augmentation errors.".format(self._nr_error))


class ImageFromFile(RNGDataFlow):
    """ Produce images read from a list of files as (h, w, c) arrays. """
    def __init__(self, files, channel=3, resize=None, shuffle=False):
        """
        Args:
            files (list): list of file paths.
            channel (int): 1 or 3. Will convert grayscale to RGB images if channel==3.
                Will produce (h, w, 1) array if channel==1.
            resize (tuple): int or (h, w) tuple. If given, resize the image.
        """
        assert len(files), "No image files given to ImageFromFile!"
        self.files = files
        self.channel = int(channel)
        assert self.channel in [1, 3], self.channel
        self.imread_mode = cv2.IMREAD_GRAYSCALE if self.channel == 1 else cv2.IMREAD_COLOR
        if resize is not None:
            resize = shape2d(resize)
        self.resize = resize
        self.shuffle = shuffle

    def __len__(self):
        return len(self.files)

    def __iter__(self):
        if self.shuffle:
            self.rng.shuffle(self.files)
        for f in self.files:
            im = cv2.imread(f, self.imread_mode)
            assert im is not None, f
            if self.channel == 3:
                im = im[:, :, ::-1]
            if self.resize is not None:
                im = cv2.resize(im, tuple(self.resize[::-1]))
            if self.channel == 1:
                im = im[:, :, np.newaxis]
            yield [im]


class AugmentImageComponent(MapDataComponent):
    """
    Apply image augmentors on 1 image component.
    """

    def __init__(self, ds, augmentors, index=0, copy=True, catch_exceptions=False):
        """
        Args:
            ds (DataFlow): input DataFlow.
            augmentors (AugmentorList): a list of :class:`imgaug.ImageAugmentor` to be applied in order.
            index (int or str): the index or key of the image component to be augmented in the datapoint.
            copy (bool): Some augmentors modify the input images. When copy is
                True, a copy will be made before any augmentors are applied,
                to keep the original images not modified.
                Turn it off to save time when you know it's OK.
            catch_exceptions (bool): when set to True, will catch
                all exceptions and only warn you when there are too many (>100).
                Can be used to ignore occasion errors in data.
        """
        if isinstance(augmentors, AugmentorList):
            self.augs = augmentors
        else:
            self.augs = AugmentorList(augmentors)
        self._copy = copy

        self._exception_handler = ExceptionHandler(catch_exceptions)
        super(AugmentImageComponent, self).__init__(ds, self._aug_mapper, index)

    def reset_state(self):
        self.ds.reset_state()
        self.augs.reset_state()

    def _aug_mapper(self, x):
        check_dtype(x)
        with self._exception_handler.catch():
            if self._copy:
                x = copy_mod.deepcopy(x)
            return self.augs.augment(x)


class AugmentImageCoordinates(MapData):
    """
    Apply image augmentors on an image and a list of coordinates.
    Coordinates must be a Nx2 floating point array, each row is (x, y).
    """

    def __init__(self, ds, augmentors, img_index=0, coords_index=1, copy=True, catch_exceptions=False):

        """
        Args:
            ds (DataFlow): input DataFlow.
            augmentors (AugmentorList): a list of :class:`imgaug.ImageAugmentor` to be applied in order.
            img_index (int or str): the index/key of the image component to be augmented.
            coords_index (int or str): the index/key of the coordinate component to be augmented.
            copy, catch_exceptions: same as in :class:`AugmentImageComponent`
        """
        if isinstance(augmentors, AugmentorList):
            self.augs = augmentors
        else:
            self.augs = AugmentorList(augmentors)

        self._img_index = img_index
        self._coords_index = coords_index
        self._copy = copy
        self._exception_handler = ExceptionHandler(catch_exceptions)

        super(AugmentImageCoordinates, self).__init__(ds, self._aug_mapper)

    def reset_state(self):
        self.ds.reset_state()
        self.augs.reset_state()

    def _aug_mapper(self, dp):
        with self._exception_handler.catch():
            img, coords = dp[self._img_index], dp[self._coords_index]
            check_dtype(img)
            validate_coords(coords)
            if self._copy:
                img, coords = copy_mod.deepcopy((img, coords))
            tfms = self.augs.get_transform(img)
            dp[self._img_index] = tfms.apply_image(img)
            dp[self._coords_index] = tfms.apply_coords(coords)
            return dp


class AugmentImageComponents(MapData):
    """
    Apply image augmentors on several components, with shared augmentation parameters.

    Example:

        .. code-block:: python

            ds = MyDataFlow()   # produce [image(HWC), segmask(HW), keypoint(Nx2)]
            ds = AugmentImageComponents(
                ds, augs,
                index=(0,1), coords_index=(2,))

    """

    def __init__(self, ds, augmentors, index=(0, 1), coords_index=(), copy=True, catch_exceptions=False):
        """
        Args:
            ds (DataFlow): input DataFlow.
            augmentors (AugmentorList): a list of :class:`imgaug.ImageAugmentor` instance to be applied in order.
            index: tuple of indices of the image components.
            coords_index: tuple of indices of the coordinates components.
            copy, catch_exceptions: same as in :class:`AugmentImageComponent`
        """
        if isinstance(augmentors, AugmentorList):
            self.augs = augmentors
        else:
            self.augs = AugmentorList(augmentors)
        self.ds = ds
        self._exception_handler = ExceptionHandler(catch_exceptions)
        self._copy = copy
        self._index = index
        self._coords_index = coords_index

        super(AugmentImageComponents, self).__init__(ds, self._aug_mapper)

    def reset_state(self):
        self.ds.reset_state()
        self.augs.reset_state()

    def _aug_mapper(self, dp):
        dp = copy_mod.copy(dp)  # always do a shallow copy, make sure the list is intact
        copy_func = copy_mod.deepcopy if self._copy else lambda x: x  # noqa
        with self._exception_handler.catch():
            major_image = self._index[0]  # image to be used to get params. TODO better design?
            im = copy_func(dp[major_image])
            check_dtype(im)
            tfms = self.augs.get_transform(im)
            dp[major_image] = tfms.apply_image(im)
            for idx in self._index[1:]:
                check_dtype(dp[idx])
                dp[idx] = tfms.apply_image(copy_func(dp[idx]))
            for idx in self._coords_index:
                coords = copy_func(dp[idx])
                validate_coords(coords)
                dp[idx] = tfms.apply_coords(coords)
            return dp


try:
    import cv2
    from .imgaug import AugmentorList
except ImportError:
    from ..utils.develop import create_dummy_class
    ImageFromFile = create_dummy_class('ImageFromFile', 'cv2')  # noqa
    AugmentImageComponent = create_dummy_class('AugmentImageComponent', 'cv2')  # noqa
    AugmentImageCoordinates = create_dummy_class('AugmentImageCoordinates', 'cv2') # noqa
    AugmentImageComponents = create_dummy_class('AugmentImageComponents', 'cv2')  # noqa



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

    tokenizer = HFTokenizer(vocab_file)
    text = "中国的英文是China，成立于1949年。"
    encodes = tokenizer.encode(text)

    tokens        : ['[CLS]', '中', '国', '的', '英', '文', '是', 'china', '，', '成', '立', '于', '1949', '年', '。', '[SEP]']
    offsets       : [(0, 0), (0, 1), (1, 2), (2, 3), (3, 4), (4, 5), (5, 6), (6, 11), (11, 12), (12, 13), (13, 14), (14, 15), (15, 19), (19, 20), (20, 21), (0, 0)]
    token2char    : [-1, 0, 1, 2, 3, 4, 5, 6, 11, 12, 13, 14, 15, 19, 20, -1]
    char2token    : [1, 2, 3, 4, 5, 6, 7, 7, 7, 7, 7, 8, 9, 10, 11, 12, 12, 12, 12, 13, 14]
    ids           : [101, 704, 1744, 4638, 5739, 3152, 3221, 8873, 8024, 2768, 4989, 754, 8594, 2399, 511, 102]
    attention_mask: [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]
    type_ids      : [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]

"""

from collections import defaultdict

from loguru import logger

from transformers import BertTokenizer


# Hugging Face Tokenizer
class HFTokenizer:
    def __init__(self,
                 vocab_file,
                 lowercase=True,
                 strip_accents=False,
                 clean_text=True,
                 cc=None):
        self.vocab_file = vocab_file
        self.cc = None
        if cc is not None:
            # pip install opencc-python-reimplemented
            import opencc
            self.cc = opencc.OpenCC(cc)
        from tokenizers import BertWordPieceTokenizer
        self._tokenizer = BertWordPieceTokenizer(self.vocab_file,
                                                 lowercase=lowercase,
                                                 strip_accents=strip_accents,
                                                 clean_text=clean_text)

    def encode(self, text, max_seq_length=None, add_special_tokens=True):
        if self.cc is not None:
            text = self.cc.convert(text)

        encodes = self._tokenizer.encode(
            text, add_special_tokens=add_special_tokens)

        tokens = encodes.tokens
        offsets = encodes.offsets
        ids = encodes.ids
        attention_mask = encodes.attention_mask
        type_ids = encodes.type_ids

        #  len(tokens) -
        #  self._tokenizer.num_special_tokens_to_add(is_pair=False))
        token2char = [-1] * (len(tokens))
        char2token = [-1] * len(text)
        for i, (start, end) in enumerate(offsets):
            if start == 0 and end == 0:
                continue
            token2char[i] = start
            for j in range(start, end):
                char2token[j] = i


        return {
            'tokens': tokens,
            'offsets': offsets,
            'token2char': token2char,
            'char2token': char2token,
            'ids': ids,
            'attention_mask': attention_mask,
            'type_ids': type_ids,
        }

    """
    texts = [
        "中国的英文是China，成立于1949年。",
        "美国的英文是American，成立于1776年。",
    ]
    batch_encodes = tokenizer.batch_encode(texts)

    return:

    tokens:  [['[CLS]', '中', '国', '的', '英', '文', '是', 'china', '，', '成', '立', '于', '1949', '年', '。', '[SEP]'], 
              ['[CLS]', '美', '国', '的', '英', '文', '是', 'american', '，', '成', '立', '于', '177', '##6', '年', '。', '[SEP]']
             ]
    offsets: [[(0, 0), (0, 1), (1, 2), (2, 3), (3, 4), (4, 5), (5, 6), (6, 11), (11, 12), (12, 13), (13, 14), (14, 15), (15, 19), (19, 20), (20, 21), (0, 0)], 
              [(0, 0), (0, 1), (1, 2), (2, 3), (3, 4), (4, 5), (5, 6), (6, 14), (14, 15), (15, 16), (16, 17), (17, 18), (18, 21), (21, 22), (22, 23), (23, 24), (0, 0)]
             ]
    token2char: [[0, 1, 2, 3, 4, 5, 6, 11, 12, 13, 14, 15, 19, 20], 
                 [0, 1, 2, 3, 4, 5, 6, 14, 15, 16, 17, 18, 21, 22, 23]
                ]
    char2token: [[1, 2, 3, 4, 5, 6, 7, 7, 7, 7, 7, 8, 9, 10, 11, 12, 12, 12, 12, 13, 14], 
                 [1, 2, 3, 4, 5, 6, 7, 7, 7, 7, 7, 7, 7, 7, 8, 9, 10, 11, 12, 12, 12, 13, 14, 15]
                ]
    ids: [[101, 704, 1744, 4638, 5739, 3152, 3221, 8873, 8024, 2768, 4989, 754, 8594, 2399, 511, 102], 
          [101, 5401, 1744, 4638, 5739, 3152, 3221, 9735, 8024, 2768, 4989, 754, 10132, 8158, 2399, 511, 102]
         ]
    attention_mask: [[1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]]
    type_ids: [[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]]
    """

    def batch_encode(self, texts, add_special_tokens=True):
        batch_encodes = defaultdict(list)
        for text in texts:
            encodes = self.encode(text, add_special_tokens=add_special_tokens)
            for k, v in encodes.items():
                batch_encodes[k].append(v)
        return batch_encodes

    def save_vocabulary(self, model_path):
        self._tokenizer.save(model_path + "/vocab.txt")


class CNerTokenizer(BertTokenizer):
    def __init__(self, vocab_file, do_lower_case=False, is_english=False):
        #  super().__init__(vocab_file=str(vocab_file),
        #                   do_lower_case=do_lower_case)
        super(CNerTokenizer, self).__init__(vocab_file=vocab_file,
                                            do_lower_case=do_lower_case)
        self.vocab_file = str(vocab_file)
        self.do_lower_case = do_lower_case
        self.is_english = is_english

    def tokenize(self, text):
        if self.is_english:
            text_tokens = []
            words = text.split(' ')
            for w in words:
                word_tokens = super(CNerTokenizer, self).tokenize(w)
            text_tokens.extend(word_tokens)
        else:
            text_tokens = [c for c in text]

        _tokens = []
        for c in text_tokens:
            if self.do_lower_case:
                c = c.lower()

            #  if super(CNerTokenizer,
            #           self).tokenize(f"{c}", add_special_tokens=True)[1:-1]:
            if c in self.vocab:
                _tokens.append(c)
            else:
                _tokens.append('[UNK]')
        return _tokens


def test(args):

    vocab_file = args.vocab_file

    tokenizer = HFTokenizer(vocab_file)

    #  text = "中国的英文是China，成立于1949年。"
    text = "中国宋代景定四年（1263）黎靖德以类编排，于咸淳二年（ 1270 ） 刊为《朱子语类大全）140卷，即今通行本《朱子语类 》"
    print(f"text: {text}")

    encodes = tokenizer.encode(text, add_special_tokens=False)
    for k, v in encodes.items():
        print(f"{k}: {v}")

    tokenizer_b = CNerTokenizer(vocab_file)
    tokens_b = tokenizer_b.tokenize(text)
    logger.warning(f"tokens_b: {tokens_b}")

    #  texts = [
    #      "中国的英文是China，成立于1949年。",
    #      "美国的英文是American，成立于1776年。",
    #  ]
    #
    #  batch_encodes = tokenizer.batch_encode(texts)
    #  for k, v in batch_encodes.items():
    #      print(f"{k}: {v}")


if __name__ == '__main__':
    from argparse import ArgumentParser
    parser = ArgumentParser()
    parser.add_argument(
        "--vocab_file",
        default="/opt/share/pretrained/pytorch/bert-base-chinese/vocab.txt",
        help="Vocab file.")
    args = parser.parse_args()

    test(args)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
from collections import Counter

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from loguru import logger
from torch.nn import BCELoss, CrossEntropyLoss
from torch.utils.data import DataLoader, RandomSampler, SequentialSampler

from transformers import (AutoConfig, AutoModelForTokenClassification,
                          BertConfig, BertTokenizer, BertTokenizerFast)
#  from transformers.modeling_bert import BertModel, BertPreTrainedModel
from transformers.models.bert import BertModel, BertPreTrainedModel

from ...losses import DiceLoss, FocalLoss, LabelSmoothingCrossEntropy
from ...utils.multiprocesses import (barrier_leader_process,
                                     barrier_member_processes,
                                     is_multi_processes)
#  from .utils import CNerTokenizer
from ..models.linears import PoolerEndLogits, PoolerStartLogits
from ..trainer import Trainer, get_default_optimizer_parameters

#  import mlflow

BertLayerNorm = torch.nn.LayerNorm


def get_active_logits(logits, num_labels):
    # Only keep active parts of the loss
    loss_sig = nn.Sigmoid()
    active_logits = logits.view(-1, num_labels * 2)
    active_logits = loss_sig(active_logits)
    active_logits = active_logits**2

    return active_logits


class BertPnForNer(BertPreTrainedModel):
    def __init__(self, config):
        super(BertPnForNer, self).__init__(config)
        self.num_labels = config.num_labels
        self.loss_type = config.loss_type
        self.focalloss_gamma = config.focalloss_gamma
        self.focalloss_alpha = config.focalloss_alpha
        self.diceloss_weight = config.diceloss_weight

        self.bert = BertModel(config)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)
        self.LayerNorm = BertLayerNorm(config.hidden_size, eps=1e-12)

        self.classifier = nn.Linear(config.hidden_size, config.num_labels * 2)

        self.init_weights()

    def forward(self,
                input_ids,
                attention_mask=None,
                token_type_ids=None,
                labels=None):
        outputs = self.bert(input_ids=input_ids,
                            attention_mask=attention_mask,
                            token_type_ids=token_type_ids)
        sequence_output = outputs[0]
        sequence_output = self.dropout(sequence_output)

        logits = self.classifier(sequence_output)

        outputs = (logits,
                   )  # add hidden states and attention if they are here

        #  # Only keep active parts of the loss
        #  loss_sig = nn.Sigmoid()
        #  active_logits = logits.view(-1, self.num_labels * 2)
        #  active_logits = loss_sig(active_logits)
        #  active_logits = active_logits**2
        active_logits = get_active_logits(logits, self.num_labels)
        if labels is not None:

            #  active_logits = get_active_logits(logits, self.num_labels)
            #  active_labels = labels.view(-1, self.num_labels * 2).float()
            active_labels = labels.reshape(-1, self.num_labels * 2).float()

            #  logger.warning(f"active_logits.shape: {active_logits.shape}")
            #  logger.warning(f"active_labels.shape: {active_labels.shape}")

            if self.loss_type == 'FocalLoss':
                loss_fct = FocalLoss(gamma=self.focalloss_gamma,
                                     alpha=self.focalloss_alpha)
                loss = loss_fct(active_logits, active_labels)
            else:
                loss_fct = BCELoss(reduction='none')
                loss = loss_fct(active_logits, active_labels)

            batch_size = input_ids.size(0)
            loss = loss.view(batch_size, -1, self.num_labels * 2)
            loss = torch.mean(loss, 2)
            loss = torch.sum(attention_mask * loss) / torch.sum(attention_mask)
            #  outputs = (loss, ) + outputs
            outputs = (loss, ) + (active_logits, )
        else:
            #  outputs = (active_logits, )
            #  outputs = (torch.tensor(0.0).cuda(), ) + outputs
            outputs = (torch.tensor(0.0).cuda(), ) + (active_logits, )

        return outputs


class SpanEntityScore(object):
    def __init__(self, id2label, ignore_categories=None):
        self.id2label = id2label
        self.ignore_categories = ignore_categories
        self.reset()

    def reset(self):
        self.origins = []
        self.founds = []
        self.rights = []

    def compute(self, origin, found, right):
        recall = 0 if origin == 0 else (right / origin)
        precision = 0 if found == 0 else (right / found)
        f1 = 0. if recall + precision == 0 else (2 * precision *
                                                 recall) / (precision + recall)
        return recall, precision, f1

    def result(self):
        class_info = {}
        origin_counter = Counter(
            [f"{x[0]}:{self.id2label[x[0]]}" for x in self.origins])
        found_counter = Counter(
            [f"{x[0]}:{self.id2label[x[0]]}" for x in self.founds])
        right_counter = Counter(
            [f"{x[0]}:{self.id2label[x[0]]}" for x in self.rights])
        total_origin = 0
        total_found = 0
        total_right = 0
        for type_, count in origin_counter.items():
            category = type_.split(':')[1]
            origin = count
            found = found_counter.get(type_, 0)
            right = right_counter.get(type_, 0)
            recall, precision, f1 = self.compute(origin, found, right)
            class_info[type_] = {
                "acc": round(precision, 4),
                'recall': round(recall, 4),
                'f1': round(f1, 4),
                'right': right,
                'found': found,
                'origin': origin
            }
            if self.ignore_categories and category in self.ignore_categories:
                pass
            else:
                total_origin += origin
                total_found += found
                total_right += right
        if self.ignore_categories:
            origin = total_origin
            found = total_found
            right = total_right
        else:
            origin = len(self.origins)
            found = len(self.founds)
            right = len(self.rights)
        recall, precision, f1 = self.compute(origin, found, right)

        return {
            'acc': precision,
            'recall': recall,
            'f1': f1,
            'right': right,
            'found': found,
            'origin': origin
        }, class_info

    def update(self, true_subject, pred_subject):
        self.origins.extend(true_subject)
        self.founds.extend(pred_subject)
        self.rights.extend([
            pre_entity for pre_entity in pred_subject
            if pre_entity in true_subject
        ])


MODEL_CLASSES = {
    #  'bert': (BertConfig, BertPnForNer, CNerTokenizer),
    'bert': (BertConfig, BertPnForNer, BertTokenizer),
    #  'bert': (BertConfig, BertSpanForNer, BertTokenizerFast),
}


def load_pretrained_tokenizer(args):
    #  config_class, model_class, tokenizer_class = MODEL_CLASSES[args.model_type]
    #  tokenizer = tokenizer_class.from_pretrained(
    #      args.model_path,
    #      do_lower_case=args.do_lower_case,
    #      is_english=args.is_english,
    #      cache_dir=args.cache_dir if args.cache_dir else None,
    #  )
    from ..token_utils import HFTokenizer
    tokenizer = HFTokenizer(os.path.join(args.model_path, 'vocab.txt'),
                            lowercase=args.do_lower_case,
                            cc=args.cc)

    return tokenizer


def load_pretrained_model(args):
    # make sure only the first process in distributed training
    # will download model & vocab
    barrier_member_processes(args)

    config_class, model_class, tokenizer_class = MODEL_CLASSES[args.model_type]
    config = config_class.from_pretrained(
        args.model_path,
        num_labels=args.num_labels,
        loss_type=args.loss_type,
        cache_dir=args.cache_dir if args.cache_dir else None,
        soft_label=args.soft_label,
    )
    setattr(config, 'label2id', args.label2id)
    setattr(config, 'id2label', args.id2label)
    setattr(config, 'loss_type', args.loss_type)
    setattr(config, 'focalloss_gamma', args.focalloss_gamma)
    setattr(config, 'focalloss_alpha', args.focalloss_alpha)
    setattr(config, 'diceloss_weight', args.diceloss_weight)
    logger.info(f"model_path: {args.model_path}")
    logger.info(f"config:{config}")
    model = model_class.from_pretrained(
        args.model_path,
        from_tf=bool(".ckpt" in args.model_path),
        config=config,
        cache_dir=args.cache_dir if args.cache_dir else None,
    )

    # make sure only the first process in distributed training
    # will download model & vocab
    barrier_leader_process(args)

    return model


def collate_fn(batch):
    """
    batch should be a list of (sequence, target, length) tuples...
    Returns a padded tensor of sequences sorted from longest to shortest,
    """
    #  from .dataset import batch_to_input_data
    #  all_input_ids, all_attention_mask, all_token_type_ids, all_labels, all_lens, all_token_offsets, all_subjects = batch_to_input_data(
    #      batch)
    all_input_ids = torch.stack([e.input_ids for e in batch])
    all_attention_mask = torch.stack([e.attention_mask for e in batch])
    all_token_type_ids = torch.stack([e.token_type_ids for e in batch])
    all_labels = torch.stack([e.labels for e in batch])
    all_input_lens = torch.stack([e.input_len for e in batch])
    all_token_offsets = torch.stack([e.token_offsets for e in batch])
    all_subjects = [e.subjects for e in batch]

    max_len = max(all_input_lens).item()
    all_input_ids = all_input_ids[:, :max_len]
    all_attention_mask = all_attention_mask[:, :max_len]
    all_token_type_ids = all_token_type_ids[:, :max_len]
    all_labels = all_labels[:, :max_len]
    all_token_offsets = all_token_offsets[:, :max_len]
    return all_input_ids, all_attention_mask, all_token_type_ids, all_labels, all_input_lens, all_token_offsets, all_subjects


def extract_entity_from_logits(args,
                               preds,
                               lens,
                               confidence=0.5,
                               enable_nested_entities=False):
    num_labels = int(preds.shape[-1] / 2)
    num_tokens = lens[0]

    #  logger.info(f"preds: (shape: {preds.shape}) | {preds}")
    preds = preds.view(1, -1, num_labels * 2)
    preds = preds.detach().cpu().numpy()

    starts = []
    ends = []
    for i in range(num_labels):
        start = np.where(preds[0, :, 2 * i] > confidence)[0]
        end = np.where(preds[0, :, 2 * i + 1] > confidence)[0]

        start = np.array([x for x in start if x >= 0 and x < num_tokens])
        end = np.array([x for x in end if x >= 0 and x < num_tokens])

        starts.append(start)
        ends.append(end)
    #  logger.info(f"starts: {starts}")
    #  logger.info(f"ends: {ends}")
    #  logger.info(f"enable_nested_entities: {enable_nested_entities}")
    entities = []
    for n in range(num_labels):
        start = starts[n]
        end = ends[n]

        last_end_pos = -1
        for idx, i in enumerate(start):
            if i <= last_end_pos:
                continue
            jj = end[end >= i]
            for j in jj:
                if idx < len(start) - 1 and j >= start[idx + 1]:
                    break
                s = i  #- 1
                e = j  #- 1
                last_end_pos = e
                entities.append((n + 1, s, e))
                if not enable_nested_entities:
                    break

    entities = sorted(entities, key=lambda x: x[1])
    return entities


def load_model(args):
    model = load_pretrained_model(args)
    model.to(args.device)
    return model


def build_default_model(args):
    """
    自定义模型
    规格要求返回模型(model)、优化器(optimizer)、调度器(scheduler)三元组。
    """

    # -------- model --------
    model = load_pretrained_model(args)
    model.to(args.device)

    # -------- optimizer --------
    from transformers.optimization import AdamW
    optimizer_parameters = get_default_optimizer_parameters(
        model, args.weight_decay)
    optimizer = AdamW(optimizer_parameters,
                      lr=args.learning_rate,
                      eps=args.adam_epsilon,
                      correct_bias=False)

    # -------- scheduler --------
    from transformers.optimization import get_linear_schedule_with_warmup

    scheduler = get_linear_schedule_with_warmup(
        optimizer,
        num_warmup_steps=args.total_steps * args.warmup_rate,
        num_training_steps=args.total_steps)

    return model, optimizer, scheduler


def init_labels(args, labels):

    args.ner_labels = labels
    args.id2label = {i + 1: label for i, label in enumerate(args.ner_labels)}
    args.label2id = {label: i + 1 for i, label in enumerate(args.ner_labels)}
    args.num_labels = len(args.label2id)

    logger.info(f"args.label2id: {args.label2id}")
    logger.info(f"args.id2label: {args.id2label}")
    logger.info(f"args.num_labels: {args.num_labels}")


class NerTrainer(Trainer):
    def __init__(self, args, ner_labels, build_model=None, tokenizer=None):
        super(NerTrainer, self).__init__(args)
        init_labels(args, ner_labels)
        if tokenizer:
            self.tokenizer = tokenizer
        else:
            self.tokenizer = load_pretrained_tokenizer(args)

        if build_model is None:
            self.build_model = build_default_model
        else:
            self.build_model = build_model

        self.label2id = args.label2id
        self.collate_fn = collate_fn

    #  def examples_to_dataset(self, examples, max_seq_length):
    #      from .dataset import examples_to_dataset
    #      return examples_to_dataset(examples, self.label2id, self.tokenizer,
    #                                 max_seq_length)
    def encode_examples(self, examples, max_seq_length):
        from .dataset import encode_examples
        return encode_examples(examples, self.label2id, self.tokenizer,
                               max_seq_length)

    def batch_to_inputs(self, args, batch, known_labels=True):
        inputs = {
            "input_ids": batch[0],
            "attention_mask": batch[1],
            "labels": batch[3],
        }
        if args.model_type != "distilbert":
            # XLM and RoBERTa don"t use token_type_ids
            inputs["token_type_ids"] = (
                batch[2] if args.model_type in ["bert", "xlnet"] else None)

        return inputs

    #  def generate_dataloader(self, args, dataset, batch_size, keep_order=True):
    #
    #      Sampler = SequentialSampler if keep_order else RandomSampler
    #      sampler = DistributedSampler(dataset) if is_multi_processes(
    #          args) else Sampler(dataset)
    #      dataloader = DataLoader(dataset,
    #                              sampler=sampler,
    #                              batch_size=batch_size,
    #                              collate_fn=collate_fn)
    #      return dataloader

    def on_eval_start(self, args, eval_features):
        self.metric = SpanEntityScore(args.id2label,
                                      ignore_categories=args.ignore_categories)
        pass

    #  def on_eval_step(self, args, eval_dataset, step, model, inputs, outputs):
    #  def on_eval_step(self, args, model, step, batch, batch_features):
    def on_eval_step(self, args, model, step, batch):
        all_input_ids, all_attention_mask, all_token_type_ids, all_labels, all_lens, all_token_offsets, all_subjects = batch

        eval_loss = 0.0
        num_eval_steps = 0
        for i in range(all_input_ids.size()[0]):
            inputs = {
                "input_ids": all_input_ids[i].view(1, -1),
                "attention_mask": all_attention_mask[i].view(1, -1),
            }
            if args.model_type != "distilbert":
                # XLM and RoBERTa don"t use token_type_ids
                inputs["token_type_ids"] = (all_token_type_ids[i].view(
                    1, -1) if args.model_type in ["bert", "xlnet"] else None)
            token_offsets = all_token_offsets[i]

            outputs = model(**inputs)
            tmp_eval_loss, logits = outputs[:2]
            eval_loss += tmp_eval_loss
            num_eval_steps += 1

            #  loss_sig = nn.Sigmoid()
            #  active_logits = logits.view(-1, args.num_labels * 2)
            #  activae_logits = loss_sig(active_logits)
            #  active_logits = active_logits**2
            #  preds = active_logits

            #  active_logits = get_active_logits(logits, args.num_labels)
            #  preds = active_logits
            preds = logits
            num_tokens = int(all_lens[i])

            T = all_subjects[i]
            R = extract_entity_from_logits(
                args,
                preds, [num_tokens],
                confidence=args.confidence,
                enable_nested_entities=args.enable_nested_entities)
            #  logger.info(f"R: {R}")
            #  logger.info(f"T: {T}")

            self.metric.update(true_subject=T, pred_subject=R)

        eval_loss = eval_loss / num_eval_steps
        eval_info, entity_info = self.metric.result()
        results = {f'{key}': value for key, value in eval_info.items()}
        results['loss'] = eval_loss

        return (eval_loss, ), results

    def on_predict_start(self, args, test_features):
        self.pred_results = []

    def on_predict_step(self, args, model, step, batch):
        all_input_ids, all_attention_mask, all_token_type_ids, all_labels, all_lens, all_token_offsets, all_subjects = batch

        for i in range(all_input_ids.size()[0]):
            inputs = {
                "input_ids": all_input_ids[i].view(1, -1),
                "attention_mask": all_attention_mask[i].view(1, -1),
            }
            if args.model_type != "distilbert":
                # XLM and RoBERTa don"t use token_type_ids
                inputs["token_type_ids"] = (all_token_type_ids[i].view(
                    1, -1) if args.model_type in ["bert", "xlnet"] else None)
            token_offsets = all_token_offsets[i]

            outputs = model(**inputs)
            logits = outputs[1]

            #  loss_sig = nn.Sigmoid()
            #  active_logits = logits.view(-1, args.num_labels * 2)
            #  activae_logits = loss_sig(active_logits)
            #  active_logits = active_logits**2
            #  preds = active_logits

            #  active_logits = get_active_logits(logits, args.num_labels)
            #  preds = active_logits
            preds = logits
            num_tokens = int(all_lens[i])

            R = extract_entity_from_logits(
                args,
                preds, [num_tokens],
                confidence=args.confidence,
                enable_nested_entities=args.enable_nested_entities)

            if R:
                #  label_entities = [[args.id2label[x[0]], x[1], x[2]] for x in R]
                label_entities = [[
                    args.id2label[x[0]], token_offsets[x[1]][0].item(),
                    token_offsets[x[2]][-1].item() - 1
                ] for x in R]
                if args.ignore_categories:
                    label_entities = [
                        x for x in label_entities
                        if x[0] not in args.ignore_categories
                    ]
                label_entities = [
                    x for x in label_entities
                    if x[1] <= x[2] and x[1] >= 0 and x[2] >= 0
                ]
            else:
                label_entities = []

            #  logger.info(f"label_entities: {label_entities}")

            #  if i < 20:
            #      logger.info(f"{i}, label_entities: {label_entities}")

            #  logger.debug(f"{label_entities}")
            json_d = {}
            json_d['id'] = step
            #  tag_seq = [args.id2label[x] for x in preds]
            #  json_d['tag_seq'] = " ".join(tag_seq)
            json_d['entities'] = label_entities

            #  logger.debug(f"{json_d}")

            self.pred_results.append(json_d)

    def on_predict_end(self, args, test_features):
        return self.pred_results

    def on_eval_end(self, args, eval_features):
        from ...utils.ner_utils import get_ner_results
        results = get_ner_results(self.metric)
        return results



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
from tqdm import tqdm
from loguru import tagger


def save_poplar_file(tagged_text_list,
                     poplar_file,
                     ner_labels,
                     ner_connections,
                     start_page=0,
                     max_pages=100):

    poplar_json = {
        "content": "",
        "labelCategories": [],
        "labels": [],
        "connectionCategories": [],
        "connections": []
    }

    poplar_colorset = [
        '#007bff', '#17a2b8', '#28a745', '#fd7e14', '#e83e8c', '#dc3545',
        '#20c997', '#ffc107', '#007bff'
    ]
    label2id = {x: i for i, x in enumerate(ner_labels)}
    label_categories = poplar_json['labelCategories']
    for _id, x in enumerate(ner_labels):
        label_categories.append({
            "id":
            _id,
            "text":
            x,
            "color":
            poplar_colorset[label2id[x] % len(poplar_colorset)],
            "borderColor":
            "#cccccc"
        })

    connection_categories = poplar_json['connectionCategories']
    for _id, _text in enumerate(ner_connections):
        connection_categories.append({'id': _id, 'text': _text})

    poplar_labels = poplar_json['labels']
    poplar_content = ""
    num_pages = 0
    page_offset = 0

    eid = 0
    for tagged_text in tqdm(tagged_text_list):
        if num_pages < start_page:
            num_pages += 1
            continue
        guid = tagged_text.guid
        text = tagged_text.text

        page_head = f"\n-------------------- {guid} Begin --------------------\n\n"
        page_tail = f"\n-------------------- {guid} End --------------------\n\n"
        poplar_content += page_head + f"{text}" + page_tail

        for entity_tag in tagged_text.tags:
            category = entity_tag.category
            mention = entity_tag.mention
            start_char = entity_tag.start
            end_char = start_char + len(mention) - 1

            poplar_labels.append({
                "id":
                eid,
                "categoryId":
                label2id[category],
                "startIndex":
                page_offset + len(page_head) + start_char,
                "endIndex":
                page_offset + len(page_head) + end_char + 1,
            })
            eid += 1

        num_pages += 1
        if num_pages - start_page >= max_pages:
            break

        page_offset = len(poplar_content)

    poplar_json["content"] = poplar_content
    poplar_json['labels'] = poplar_labels

    json.dump(poplar_json,
              open(poplar_file, 'w'),
              ensure_ascii=False,
              indent=2)
    logger.info(f"Saved {poplar_file}")


def poplar_data_generator(poplar_file):
    tagged_train_json_data = json.load(open(poplar_file, 'r'))

    all_labels = tagged_train_json_data['labelCategories']
    id2label = {x['id']: x['text'] for x in all_labels}

    all_entities = tagged_train_json_data['labels']

    content = tagged_train_json_data['content']

    #  re_b = '(\\n[-]+ yanbao\\d\\d\\d\\.txt Begin [-]+\\n\\n)'
    #  re_e = '(\\n[-]+ yanbao\\d\\d\\d\\.txt End [-]+\\n\\n)'
    re_b = '(\\n[-]+ (.*?) Begin [-]+\\n\\n)'
    re_e = '(\\n[-]+ (.*?) End [-]+\\n\\n)'

    b_list = []
    for x in re.finditer(re_b, content):
        b_list.append((x.start(), x.end(), x.groups()[1]))
    e_list = []
    for x in re.finditer(re_e, content):
        e_list.append((x.start(), x.end()))

    pages = [(x_b[2], x_b[0], x_b[1], x_e[0], x_e[1])
             for x_b, x_e in zip(b_list, e_list)]

    logger.warning(f"pages: {pages}")

    for i, page in enumerate(pages):
        guid, head_x0, head_x1, tail_x0, tail_x1 = page

        text = content[head_x1:tail_x0]

        json_tags = []
        for entity in all_entities:
            s = entity['startIndex']
            e = entity['endIndex'] - 1
            m = text[s:e + 1]
            assert e >= s
            if s >= head_x1 and e < tail_x0:
                json_tags.append({
                    'category': id2label[entity['categoryId']],
                    'start': s - head_x1,
                    'mention': m,
                })
        yield guid, text, None, json_tags



from transformers import BertTokenizer
import torch.nn as nn
import torch.utils.data as Data
import json
import torch
import numpy as np
from tqdm import tqdm
from REmodel import REModel_sbuject_2, REModel_sbuject_1
from transformers import AdamW, WarmupLinearSchedule
from torch.optim.lr_scheduler import LambdaLR
from bert4keras.tokenizers import Tokenizer
import os
os.environ["CUDA_VISIBLE_DEVICES"] = "3"
model_path = '/opt/share/pretrained/pytorch/bert-base-chinese'
import random
random.seed(42)
maxlen = 256
batch_size = 40
best_acc = 0.0
output_dir = './outputs'


def load_data(filename):
    D = []
    with open(filename, 'r', encoding='utf8') as f:
        for l in f:
            l = json.loads(l)
            d = {'text': l['text'], 'spo_list': []}
            for spo in l['spo_list']:
                for k, v in spo['object'].items():
                    d['spo_list'].append(
                        (spo['subject'], spo['predicate'] + '_' + k, v))
            D.append(d)
    return D


#加载数据集
train_data = load_data('/home/ycy/HBT/data/train_data.json')
valid_data = load_data('/home/ycy/HBT/data/dev_data.json')

#读取schema
with open('/home/ycy/HBT/data/schema.json', encoding='utf8') as f:
    id2predicate, predicate2id, n = {}, {}, 0
    predicate2type = {}
    for l in f:
        l = json.loads(l)
        predicate2type[l['predicate']] = (l['subject_type'], l['object_type'])
        for k, _ in sorted(l['object_type'].items()):
            key = l['predicate'] + '_' + k
            id2predicate[n] = key
            predicate2id[key] = n
            n += 1
# tokenizer = OurTokenizer(vocab_file=BERT_PATH + "vocab.txt")
tokenizer = BertTokenizer.from_pretrained(model_path, do_lower=True)
tokenizer_k = Tokenizer(os.path.join(model_path, 'vocab.txt'),
                        do_lower_case=True)


def search(pattern, sequence):
    """从sequence中寻找子串pattern
    如果找到，返回第一个下标；否则返回-1。
    """
    n = len(pattern)
    for i in range(len(sequence)):
        if sequence[i:i + n] == pattern:
            return i
    return -1


def sequence_padding(inputs, length=None, padding=0):
    """Numpy函数，将序列padding到同一长度
    """
    if length is None:
        length = max([len(x) for x in inputs])

    outputs = np.array([
        np.concatenate([x, [padding] *
                        (length - len(x))]) if len(x) < length else x[:length]
        for x in inputs
    ])
    return outputs


def combine_spoes(spoes):
    """合并SPO成官方格式
    """
    new_spoes = {}
    for s, p, o in spoes:
        p1, p2 = p.split('_')
        if (s, p1) in new_spoes:
            new_spoes[(s, p1)][p2] = o
        else:
            new_spoes[(s, p1)] = {p2: o}

    return [(k[0], k[1], v) for k, v in new_spoes.items()]


class SPO(tuple):
    """用来存三元组的类
    表现跟tuple基本一致，只是重写了 __hash__ 和 __eq__ 方法，
    使得在判断两个三元组是否等价时容错性更好。
    """
    def __init__(self, spo):
        self.spox = (
            tuple(tokenizer_k.tokenize(spo[0])),
            spo[1],
            tuple(
                sorted([(k, tuple(tokenizer_k.tokenize(v)))
                        for k, v in spo[2].items()])),
        )

    def __hash__(self):
        return self.spox.__hash__()

    def __eq__(self, spo):
        return self.spox == spo.spox


def extract_spoes(text):
    """抽取输入text所包含的三元组
    """
    tokens = tokenizer_k.tokenize(text, max_length=maxlen)
    # tokens_1 = tokenizer.tokenize(text,max_length=maxlen)
    mapping = tokenizer_k.rematch(text, tokens)
    token_ids, segment_ids = tokenizer_k.encode(text, max_length=maxlen)
    # token_ids_1 = tokenizer.encode_plus(text,max_length=maxlen)['input_ids']
    token_ids_1 = token_ids
    segment_ids_1 = segment_ids
    token_ids = torch.tensor([token_ids]).to(device)
    segment_ids = torch.tensor([segment_ids]).to(device)

    logger.info(f"subject: token_ids.shape: {token_ids.shape}")
    logger.info(f"subject: segment_ids.shape: {segment_ids.shape}")
    subject_preds = sub_model(input_ids=token_ids,
                              token_type_ids=segment_ids,
                              batch_size=1,
                              sub_train=True)

    # 抽取subject
    subject_preds = subject_preds.view(1, -1, 2)
    subject_preds = subject_preds.detach().cpu().numpy()
    # print(subject_preds)
    start = np.where(subject_preds[0, :, 0] > 0.3)[0]
    end = np.where(subject_preds[0, :, 1] > 0.3)[0]
    subjects = []
    for i in start:
        j = end[end >= i]
        if len(j) > 0:
            j = j[0]
            subjects.append((i, j))
    if subjects:
        spoes = []
        token_ids = np.repeat([token_ids_1], len(subjects), 0)
        segment_ids = np.repeat([segment_ids_1], len(subjects), 0)
        segment_ids = torch.tensor(segment_ids).to(device)
        token_ids = torch.tensor(token_ids).to(device)
        subjects_len = len(subjects)
        subjects = torch.tensor(subjects).to(device)

        logger.warning(f"subjects: {subjects}")
        logger.info(
            f"token_ids.shape: {token_ids.shape}, segment_ids.shape: {segment_ids.shape}"
        )

        # 传入subject，抽取object和predicate
        subject_preds, object_preds = sub_model(input_ids=token_ids,
                                                token_type_ids=segment_ids,
                                                subject_ids=subjects,
                                                batch_size=subjects_len,
                                                sub_train=True,
                                                obj_train=True)
        object_preds = object_preds.detach().cpu().numpy()
        for subject, object_pred in zip(subjects, object_preds):
            start = np.where(object_pred[:, :, 0] > 0.3)
            end = np.where(object_pred[:, :, 1] > 0.3)
            for _start, predicate1 in zip(*start):
                for _end, predicate2 in zip(*end):
                    if _start <= _end and predicate1 == predicate2:
                        if len(mapping[subject[0]]) > 0 and len(
                                mapping[subject[1]]) > 0 and len(
                                    mapping[_start]) > 0 and len(
                                        mapping[_end]) > 0:
                            spoes.append(
                                ((mapping[subject[0]][0],
                                  mapping[subject[1]][-1]), predicate1,
                                 (mapping[_start][0], mapping[_end][-1])))
                        break
        return [(text[s[0]:s[1] + 1], id2predicate[p], text[o[0]:o[1] + 1])
                for s, p, o, in spoes]
    else:
        return []


def evaluate(data):
    """评估函数，计算f1、precision、recall
    """
    X, Y, Z = 1e-10, 1e-10, 1e-10
    f = open('dev_pred_2.json', 'w', encoding='utf-8')
    pbar = tqdm()
    for d in data:

        R = combine_spoes(extract_spoes(d['text']))
        T = combine_spoes(d['spo_list'])
        R = set([SPO(spo) for spo in R])
        T = set([SPO(spo) for spo in T])
        X += len(R & T)
        Y += len(R)
        Z += len(T)
        f1, precision, recall = 2 * X / (Y + Z), X / Y, X / Z
        pbar.update()
        pbar.set_description('f1: %.5f, precision: %.5f, recall: %.5f' %
                             (f1, precision, recall))
        s = json.dumps(
            {
                'text': d['text'],
                'spo_list': list(T),
                'spo_list_pred': list(R),
                'new': list(R - T),
                'lack': list(T - R),
            },
            ensure_ascii=False,
            indent=4)
        f.write(s + '\n')
    pbar.close()
    f.close()
    return f1, precision, recall


class data_generator:
    """数据生成器
    """
    def __init__(self, data, batch_size=batch_size, buffer_size=None):
        self.data = data
        self.batch_size = batch_size
        if hasattr(self.data, '__len__'):
            self.steps = len(self.data) // self.batch_size
            if len(self.data) % self.batch_size != 0:
                self.steps += 1
        else:
            self.steps = None
        self.buffer_size = buffer_size or batch_size * 1000

    def __len__(self):
        return self.steps

    def data_res(self):
        batch_token_ids, batch_segment_ids, batch_attention_mask = [], [], []
        batch_subject_labels, batch_subject_ids, batch_object_labels = [], [], []
        indices = list(range(len(self.data)))
        # print(len(self.data))
        np.random.shuffle(indices)
        for i in indices:
            d = self.data[i]
            token = tokenizer.encode_plus(d['text'], max_length=maxlen)
            token_ids, segment_ids, attention_mask = token['input_ids'], token[
                'token_type_ids'], token['attention_mask']
            # 整理三元组 {s: [(o, p)]}
            spoes = {}
            for s, p, o in d['spo_list']:
                s = tokenizer.encode_plus(s)['input_ids'][1:-1]
                p = predicate2id[p]
                o = tokenizer.encode_plus(o)['input_ids'][1:-1]
                s_idx = search(s, token_ids)
                o_idx = search(o, token_ids)
                if s_idx != -1 and o_idx != -1:
                    s = (s_idx, s_idx + len(s) - 1)
                    o = (o_idx, o_idx + len(o) - 1, p)
                    if s not in spoes:
                        spoes[s] = []
                    spoes[s].append(o)
            if spoes:
                # subject标签
                subject_labels = np.zeros((len(token_ids), 2))
                for s in spoes:
                    subject_labels[s[0], 0] = 1
                    subject_labels[s[1], 1] = 1
                # 随机选一个subject
                start, end = np.array(list(spoes.keys())).T
                start = np.random.choice(start)
                end = np.random.choice(end[end >= start])
                subject_ids = (start, end)
                # 对应的object标签
                object_labels = np.zeros(
                    (len(token_ids), len(predicate2id), 2))
                for o in spoes.get(subject_ids, []):
                    object_labels[o[0], o[2], 0] = 1
                    object_labels[o[1], o[2], 1] = 1
                # 构建batch
                batch_token_ids.append(token_ids)
                batch_segment_ids.append(segment_ids)
                batch_subject_labels.append(subject_labels)
                batch_subject_ids.append(subject_ids)
                batch_object_labels.append(object_labels)
                batch_attention_mask.append(attention_mask)
        batch_token_ids = sequence_padding(batch_token_ids)
        batch_segment_ids = sequence_padding(batch_segment_ids)
        batch_subject_labels = sequence_padding(batch_subject_labels,
                                                padding=np.zeros(2))
        batch_subject_ids = np.array(batch_subject_ids)
        batch_object_labels = sequence_padding(batch_object_labels,
                                               padding=np.zeros(
                                                   (len(predicate2id), 2)))
        batch_attention_mask = sequence_padding(batch_attention_mask)
        return [
            batch_token_ids, batch_segment_ids, batch_subject_labels,
            batch_subject_ids, batch_object_labels, batch_attention_mask
        ]


class Dataset(Data.Dataset):
    def __init__(self, _batch_token_ids, _batch_segment_ids,
                 _batch_subject_labels, _batch_subject_ids,
                 _batch_obejct_labels, _batch_attention_mask):
        self.batch_token_data_ids = _batch_token_ids
        self.batch_segment_data_ids = _batch_segment_ids
        self.batch_subject_data_labels = _batch_subject_labels
        self.batch_subject_data_ids = _batch_subject_ids
        self.batch_object_data_labels = _batch_obejct_labels
        self.batch_attention_mask = _batch_attention_mask
        self.len = len(self.batch_token_data_ids)

    def __getitem__(self, index):
        return self.batch_token_data_ids[index],self.batch_segment_data_ids[index],self.batch_subject_data_labels[index],\
               self.batch_subject_data_ids[index],self.batch_object_data_labels[index],self.batch_attention_mask[index]

    def __len__(self):
        return self.len


def collate_fn(data):
    batch_token_ids = np.array([item[0] for item in data], np.int32)
    batch_segment_ids = np.array([item[1] for item in data], np.int32)
    batch_subject_labels = np.array([item[2] for item in data], np.int32)
    batch_subject_ids = np.array([item[3] for item in data], np.int32)
    batch_object_labels = np.array([item[4] for item in data], np.int32)
    batch_attention_mask = np.array([item[5] for item in data], np.int32)
    return {
        'batch_token_ids': torch.LongTensor(batch_token_ids),  # targets_i
        'batch_segment_ids': torch.FloatTensor(batch_segment_ids),
        'batch_subject_labels': torch.FloatTensor(batch_subject_labels),
        'batch_subject_ids': torch.LongTensor(batch_subject_ids),
        'batch_object_labels': torch.LongTensor(batch_object_labels),
        'batch_attention_mask': torch.LongTensor(batch_attention_mask)
    }


# def accuracy(out, labels):
#     outputs = np.argmax(out, axis=1) # 输出最大值
#     return f1_score(labels, outputs, labels=[0, 1], average='macro') # 分别做 f1 并取平均

dg = data_generator(train_data)
dg_dev = data_generator(valid_data)
T, S1, S2, K1, K2, M1 = dg.data_res()
torch_dataset = Dataset(T, S1, S2, K1, K2, M1)
loader_train = Data.DataLoader(
    dataset=torch_dataset,  # torch TensorDataset format
    batch_size=batch_size,  # mini batch size
    shuffle=True,  # random shuffle for training
    num_workers=32,
    collate_fn=collate_fn,  # subprocesses for loading data
)
model_name_or_path = '/home/ycy/roberta_zh'
sub_model = REModel_sbuject_2.from_pretrained(model_name_or_path,
                                              num_labels=2,
                                              output_hidden_states=True)
no_cuda = False
device = torch.device(
    "cuda" if torch.cuda.is_available() and not no_cuda else "cpu")
weight_decay = 0.01
learning_rate = 4e-05
adam_epsilon = 1e-05
warmup_steps = 0
epochs = 20
fp16 = False
if fp16 == True:
    sub_model.half()
# sub_model.to(device)
f = open(r'res.txt', 'a', encoding='utf8')
param_optimizer = list(sub_model.named_parameters())  # 打印每一次 迭代元素的名字与参数
# batch_token_ids = loader_res['batch_token_ids'].cuda()
# print(batch_token_ids)
#     # hack to remove pooler, which is not used
#     # thus it produce None grad that break apex
no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
optimizer_grouped_parameters = [
    {
        'params':
        [p for n, p in param_optimizer if not any(nd in n for nd in no_decay)],
        'weight_decay':
        weight_decay
    },  # n wei 层的名称, p为参数
    {
        'params':
        [p for n, p in param_optimizer if any(nd in n for nd in no_decay)],
        'weight_decay':
        0.0
    }
    # 如果是 no_decay 中的元素则衰减为 0
]
#
optimizer = AdamW(optimizer_grouped_parameters,
                  lr=learning_rate,
                  eps=adam_epsilon)  # adamw算法
# scheduler = LambdaLR(optimizer, lr_lambda=lambda epoch: 1 / (1 + 0.05 * epoch))
sub_model.to(device)
sub_model.train()
train_loss = 0.0
for epoch in range(epochs):
    for setp, loader_res in tqdm(iter(enumerate(loader_train))):
        # scheduler = WarmupLinearSchedule(optimizer, warmup_steps=warmup_steps,
        #                                  t_total=train_steps)  # warmup can su
        batch_token_ids = loader_res['batch_token_ids'].to(device)
        batch_segment_ids = loader_res['batch_segment_ids'].to(device)
        batch_subject_labels = loader_res['batch_subject_labels'].long().to(
            device)
        batch_subject_ids = loader_res['batch_subject_ids'].to(device)
        batch_object_labels = loader_res['batch_object_labels'].to(device)
        labels_start = (batch_subject_labels[:, :, 0].to(device))
        labels_end = (batch_subject_labels[:, :, 1].to(device))
        batch_attention_mask = loader_res['batch_attention_mask'].long().to(
            device)
        batch_segment_ids = batch_segment_ids.long().to(device)
        batch_attention_mask = batch_attention_mask.long().to(device)
        sub_out, obj_out = sub_model(input_ids=batch_token_ids,
                                     token_type_ids=batch_segment_ids,
                                     attention_mask=batch_attention_mask,
                                     labels=batch_subject_labels,
                                     subject_ids=batch_subject_ids,
                                     batch_size=batch_token_ids.size()[0],
                                     obj_labels=batch_object_labels,
                                     sub_train=True,
                                     obj_train=True)
        obj_loss, scores = obj_out[0:2]
        nn.utils.clip_grad_norm_(parameters=sub_model.parameters(), max_norm=1)
        obj_loss.backward()
        # print(sub_loss)
        # print(obj_loss.item())
        train_loss += obj_loss.item()
        train_loss = round(train_loss, 4)
        optimizer.step()
        # scheduler.step()
        optimizer.zero_grad()
    sub_model.eval()
    f1, precision, recall = evaluate(valid_data)
    if f1 > best_acc:
        print("Best F1", f1)
        print("Saving Model......")
        best_acc = f1
        # Save a trained model
        model_to_save = sub_model.module if hasattr(
            sub_model, 'module') else sub_model  # Only save the model it-self
        output_model_file = os.path.join(output_dir, "pytorch_model.bin")
        torch.save(model_to_save.state_dict(), output_model_file)  # 仅保存学习到的参数
        f.write(
            str(epoch) + '\t' + str(f1) + '\t' + str(precision) + '\t' +
            str(recall) + '\n')
    print(f1, precision, recall)
    f.write(
        str(epoch) + '\t' + str(f1) + '\t' + str(precision) + '\t' +
        str(recall) + '\n')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os, random, json
import numpy as np
from tqdm import tqdm
from loguru import logger
import math
import torch
from sklearn.metrics import f1_score
from .multiprocesses import is_single_process
from typing import List
from copy import deepcopy


class DataClassBase:
    def to_dict(self):
        #  return self.__dict__
        dict_obj = {}
        for k, v in self.__dict__.items():
            if k.startswith('_'):
                continue
            if isinstance(v, List):
                dict_obj[k] = [
                    x.to_dict() if isinstance(x, DataClassBase) else x
                    for x in v
                ]
            else:
                dict_obj[k] = v
        return dict_obj

    def from_dict(self, dict_data):
        for k in self.__dict__.keys():
            if k in dict_data:
                v = dict_data[k]
                setattr(self, k, v)


def generate_token_offsets(text, tokens):
    """
    参数：
        tokens是text被tokenizer切分后的词序列
    输出：
        tok2txt: {token列表序号: token在text中的偏移位置}
    """
    token_offsets = []

    text = text.lower()
    p0 = 0
    #  logger.info(f"text: {text}, tokens: {tokens}")
    for i, w in enumerate(tokens):
        #  logger.warning(f"w: {w}")
        assert len(w) > 0
        if w == '[UNK]':
            token_offsets.append(p0)
            continue
        #  if w.startswith("##"):
        #      w = w[2:]
        w = w.lower()
        #  logger.info(f"find w: {w} in | {text[p0:]}")
        p1 = text.find(w, p0)
        #  logger.info(f"p1: {p1}, p0: {p0}")
        assert p1 >= p0

        token_offsets.append(p1)
        p0 = p1 + len(w)

    assert len(token_offsets) == len(tokens)
    return token_offsets


def get_token_index(token_offsets, offset):
    from bisect import bisect_right
    index = bisect_right(token_offsets, offset) - 1
    assert index >= 0 and index < len(token_offsets)
    return index


# 英文文本切分成词序列，原始标注是基于字符，需要转换成token位置
def tokenize_en(tokenizer, text, do_lower_case=False):
    """
    返回
    text_tokens: [word1, word2, ...]
    tok2txt: {token序号: token在text的起始位置}
    """
    text_tokens = []
    words = text.split(' ')
    for w in words:
        toks = tokenizer.tokenize(w)

        # 合并被切开的"##"开头的片断
        new_toks = []
        for tok in toks:
            if not tok.startswith('##'):
                new_toks.append(tok)
            else:
                new_toks[-1] += tok[2:]
        toks = new_toks

        text_tokens.extend(toks)

    token_offsets = generate_token_offsets(text, text_tokens)

    return text_tokens, token_offsets


# 中文文本切分词序列
def tokenize_cn(tokenizer, text, do_lower_case=False):
    """
    返回
    text_tokens: [token1, token2, ...]
    tok2txt: {token序号: token在text的起始位置}
    """

    text_tokens = []
    for c in text_tokens:
        if do_lower_case:
            c = c.lower()

        #  if super(CNerTokenizer,
        #           self).tokenize(f"{c}", add_special_tokens=True)[1:-1]:
        if c in self.vocab:
            text_tokens.append(c)
        else:
            text_tokens.append('[UNK]')

    tok2txt = {}
    p0 = 0
    words = [tokenizer.decode(tok) for tok in text_tokens]
    for i, w in enumerate(words):
        assert len(w) > 0
        p1 = text.find(w, p0)
        assert p1 >= p0
        offsets.append(p1)
        tok2txt[i] = p1
        p0 = p1 + len(w)

    return text_tokens, tok2txt


def load_pytorch_model(model, model_path):
    if hasattr(model, "module"):
        model.module.load_state_dict(torch.load(model_path,
                                                map_location='cpu'),
                                     strict=False)
    else:
        model.load_state_dict(torch.load(model_path, map_location='cpu'),
                              strict=False)
    return model


def softmax(x):
    e_x = np.exp(x - np.max(x, axis=1).reshape(-1, 1))
    return e_x / e_x.sum(axis=1).reshape(-1, 1)


def sigmoid(x):
    return 1 / (1 + np.exp(-x))


def simple_accuracy(preds, labels):
    return (preds == labels).mean()


def acc_and_f1(preds, labels):
    correct = np.sum((preds == labels).astype(int))
    acc = correct / preds.shape[0]
    f1 = f1_score(y_true=labels, y_pred=preds, average="macro")
    acc_and_f1 = (acc + f1) / 2
    return {"acc": acc, "f1": f1, "acc_and_f1": acc_and_f1}


def seed_everything(seed):
    import os, random
    random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False


def init_random(seed):

    random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    np.random.seed(seed)

    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.deterministic = True


# Setup CUDA, GPU & distributed training
def init_cuda(args):
    #  if args.local_rank == -1 or args.no_cuda:
    if is_single_process(args) or args.no_cuda:
        device = torch.device("cuda" if torch.cuda.is_available()
                              and not args.no_cuda else "cpu")
        args.n_gpu = torch.cuda.device_count()
    else:
        # Initializes the distributed backend which will
        # take care of sychronizing nodes/GPUs
        torch.cuda.set_device(args.local_rank)
        device = torch.device("cuda", args.local_rank)
        torch.distributed.init_process_group(backend="nccl")
        args.n_gpu = 1
    args.train_batch_size = args.per_gpu_train_batch_size * max(1, args.n_gpu)
    args.eval_batch_size = args.per_gpu_eval_batch_size * max(1, args.n_gpu)
    args.device = device


def init_theta(args):
    #  init_random(args.seed)
    seed_everything(args.seed)
    init_cuda(args)


# DataFrame直接初始化出InputExample列表
def dataframe_to_examples(df_data, InputExample, cols_map=None):
    if cols_map is None:
        cols_map = {x: x for x in df_data.columns()}
    examples = []
    for i, row in df_data.iterrows():
        values = {k: str(row[v]) for k, v in cols_map.items()}
        examples.append(InputExample(**values))

    return np.array(examples)


def get_list_size(samples: [list, np.array]):
    if isinstance(samples, list):
        num_samples = len(samples)
    elif isinstance(samples, np.ndarray):
        num_samples = samples.shape[0]
    else:
        raise TypeError(
            f"Argument samples in shuffle_list() must be a list or np.ndarray."
        )
    return num_samples


def shuffle_list(samples: [list, np.array],
                 random_state=None) -> [list, np.array]:
    if random_state:
        np.random.seed(random_state)

    num_samples = get_list_size(samples)
    indices = np.random.randint(0, num_samples, num_samples)
    shuffled_samples = [samples[i] for i in indices]

    if isinstance(samples, np.ndarray):
        return np.array(shuffled_samples)
    else:
        return shuffled_samples


def list_to_list(samples: [list, np.array]) -> list:
    if isinstance(samples, np.ndarray):
        samples = samples.tolist()
        return samples
    elif isinstance(samples, list):
        return samples
    else:
        raise TypeError(
            f"Samples in list_to_list() must be a list or np.ndarray.")


def concatenate_list(a_samples: [list, np.array],
                     b_samples: [list, np.array]) -> [list, np.array]:
    if isinstance(a_samples, np.ndarray):
        a_samples = list_to_list(a_samples)
        b_samples = list_to_list(b_samples)
        samples = a_samples + b_samples
        return np.array(samples)
    elif isinstance(a_samples, list):
        a_samples = list_to_list(a_samples)
        b_samples = list_to_list(b_samples)
        samples = a_samples + b_samples
        return samples
    else:
        raise TypeError(
            f"Samples in concatenote_list() must be a list or np.ndarray.")


def to_numpy(X):
    """
    Convert input to numpy ndarray
    """
    if hasattr(X, 'iloc'):  # pandas
        return X.values
    elif isinstance(X, list):  # list
        return np.array(X)
    elif isinstance(X, np.ndarray):  # ndarray
        return X
    else:
        raise ValueError("Unable to handle input type %s" % str(type(X)))


def unpack_text_pairs(X):
    assert isinstance(X, list) or isinstance(X, np.ndarray)
    return (X, None) if X.ndim == 1 else (X[:, 0], X[:, 1])


def unpack_data(X, y=None):
    return unpack_text_pairs(to_numpy(X)), to_numpy(y) if y else None


def create_logger(log_dir: str, logger_name: str):
    # 创建一个logger
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.DEBUG)

    if not os.path.exists(log_dir):
        os.mkdir(log_dir)

    # 创建一个handler，
    timestamp = time.strftime("%Y.%m.%d_%H.%M.%S", time.localtime())
    log_filename = f"{log_dir}/log_{logger_name}_{timestamp}.txt"
    fh = logging.FileHandler(log_filename)
    fh.setLevel(logging.DEBUG)

    # 再创建一个handler，用于输出到控制台
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)

    # 定义handler的输出格式
    formatter = logging.Formatter(
        '[%(asctime)s][%(levelname)s] ## %(message)s')
    fh.setFormatter(formatter)
    ch.setFormatter(formatter)
    # 给logger添加handler
    logger.addHandler(fh)
    logger.addHandler(ch)

    return logger


def seg_generator(iterables, seg_len, seg_backoff=0):
    if seg_len <= 0:
        yield iterables, 0
    else:
        #  # 确保iterables列表中每一项的条目数相同
        #  assert sum([len(x)
        #              for x in iterables]) == len(iterables[0]) * len(iterables)
        assert iterables[0] is not None
        s0 = 0
        while s0 < len(iterables[0]):
            s1 = s0 + seg_len
            segs = [x[s0:s1] if x else None for x in iterables]
            yield segs, s0
            s0 += seg_len - seg_backoff


class slide_generator:
    def __init__(self, words: list, slide_len, slide_backoff=0):

        assert slide_offset <= slide_len

        self.words = words
        self.slide_len = slide_len
        self.slide_backoff = slide_backoff

        num_words = len(words)
        self.total_slides = num_words // (slide_len - slide_backoff) + 1

    def __iter__(self):
        for i in range(self.total_slides):
            yield self.__getitem__(i)

    def __len__(self):
        return self.total_slides

    def __getitem__(self, idx):
        s = (self.slide_len - self.slide_backoff) * idx
        e = (self.slide_len - self.slide_backoff) * (idx + 1)
        slide_words = self.words[s:e]

        return slide_words


def split_train_eval_examples(examples: [list, np.array],
                              train_rate=0.9,
                              fold=0,
                              shuffle=False,
                              random_type=None):
    if train_rate == 0.0:
        return [], examples
    if train_rate == 1.0:
        return examples, []
    assert train_rate > 0.0 and train_rate < 1.0

    examples = list(examples)
    num_examples = len(examples)

    # for medical_entity
    if random_type == "np":
        if shuffle:
            examples = shuffle_list(examples, random_state=None)

        num_eval_examples = int(num_examples * (1 - train_rate))
        assert fold <= num_examples // num_eval_examples

        s = num_eval_examples * fold
        e = num_eval_examples * (fold + 1)
        logger.warning(f"Eval examples: [{s}:{e}]")
        eval_examples = examples[s:e]
        train_examples = concatenate_list(examples[:s], examples[e:])
    else:
        # for entity_typing
        if shuffle:
            random.shuffle(examples)

        num_train_examples = int(num_examples * train_rate)
        num_eval_examples = num_examples - num_train_examples
        logger.info(
            f"num_train_examples: {num_train_examples}, num_eval_examples: {num_eval_examples}"
        )

        e = num_examples - num_eval_examples * fold
        s = num_examples - num_eval_examples * (fold + 1)
        logger.warning(f"split examples fold={fold} | s: {s}, e: {e}")
        if s < 0:
            s = 0
        logger.warning(f"Eval examples: [{s}:{e}]")
        eval_examples = examples[s:e]
        train_examples = concatenate_list(examples[:s], examples[e:])

    return train_examples, eval_examples


class batch_generator:
    def __init__(self,
                 examples: [list, np.array],
                 train_rate=0.9,
                 shuffle=False,
                 random_state=None):
        if shuffle:
            examples = shuffle_list(examples, random_state=random_state)
        num_examples = get_list_size(examples)
        num_eval_examples = int(num_examples * (1 - train_rate))

        self.examples = examples
        self.total_batchs = num_examples // num_eval_examples + 1

    def __iter__(self):
        for i in range(self.total_batchs):
            yield self.__getitem__(i)

    def __getitem__(self, idx):
        s = self.num_eval_examples * idx
        e = self.num_eval_examples * (idx + 1)
        eval_examples = self.examples[s:e]
        train_examples = concatenate_list(self.examples[:s], self.examples[e:])
        return train_examples, eval_examples

    def __len__(self):
        return self.total_batchs


def get_pred_results_file(args):
    return f"{args.output_dir}/{args.dataset_name}_result.json"


def get_submission_file(args):
    return f"{args.output_dir}/{args.dataset_name}_submission.json"


def load_json_file(json_file):
    rd = open(json_file, 'r')
    lines = rd.readlines()
    rd.close()
    json_data = []
    for line in tqdm(lines):
        line = line.strip()
        line_data = json.loads(line)
        json_data.append(line_data)
    print(f"Total: {len(json_data)}")
    print(json_data[:5])
    return json_data


def show_statistical_distribution(data_list):
    logger.info(f"***** Statistical Distribution *****")
    logger.info(f"mean: {np.mean(data_list):.4f}")
    logger.info(f"std: {np.mean(data_list):.4f}")
    logger.info(f"max: {np.max(data_list):.4f}")
    logger.info(f"min: {np.min(data_list)}:.4f")


def remove_duplicate_entities(R):
    new_R = []
    for i, r in enumerate(R):
        found_dup = False
        for r1 in R[i + 1:]:
            if r == r1:
                found_dup = True
                break
        if found_dup:
            continue
        new_R.append(r)
    return new_R


def is_uniform(entities):
    for i, x in enumerate(entities):
        for x1 in entities[i:]:
            if x != x1:
                return False
    return True


def merge_entities(entities_list, key=None, min_dups=2):
    new_X = []
    X = [x for z in entities_list for x in z]
    #  logger.info(f"X: {X}")
    if key:
        X = sorted(X, key=key)

    for i, x in enumerate(X[:1 - min_dups]):
        if is_uniform(X[i:i + min_dups]):
            new_X.append(deepcopy(x))
    new_X = remove_duplicate_entities(new_X)
    return new_X


def match_tokenized_to_untokenized(subwords, sentence):
    token_subwords = np.zeros(len(sentence))

    def _run_strip_accents(text):
        """Strips accents from a piece of text."""
        import unicodedata
        text = unicodedata.normalize("NFD", text)
        output = []
        for char in text:
            cat = unicodedata.category(char)
            if cat == "Mn":
                continue
            output.append(char)
        return "".join(output)

    sentence = [_run_strip_accents(x) for x in sentence]

    token_ids, subwords_str, current_token, current_token_normalized = [
        -1
    ] * len(subwords), "", 0, None
    logger.warning(f"subwords: {subwords}")
    for i, subword in enumerate(subwords):
        if subword in ["[CLS]", "[SEP]"]: continue

        while current_token_normalized is None:
            current_token_normalized = sentence[current_token].lower()

        if subword.startswith("[UNK]"):
            #  unk_length = int(subword[6:])
            unk_length = 1
            subwords[i] = subword[:5]
            subwords_str += current_token_normalized[len(subwords_str
                                                         ):len(subwords_str) +
                                                     unk_length]
        else:
            subwords_str += subword[2:] if subword.startswith(
                "##") else subword
        if not current_token_normalized.startswith(subwords_str):
            return False

        token_ids[i] = current_token
        token_subwords[current_token] += 1
        if current_token_normalized == subwords_str:
            subwords_str = ""
            current_token += 1
            current_token_normalized = None

    logger.info(f"current_token: {current_token}, sentence: {sentence}")
    logger.info(f"subwords: {subwords}")
    logger.info(f"token_ids: {token_ids}")
    assert current_token_normalized is None
    while current_token < len(sentence):
        assert not sentence[current_token]
        current_token += 1
    assert current_token == len(sentence)

    return token_ids
